from typing import TYPE_CHECKING, Any, Dict, List, Tuple, cast
from uuid import uuid4

import pydash
from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import (
    AwaitingAsyncProcessing,
    FidesopsException,
    PrivacyRequestError,
)
from fides.api.models.policy import Policy
from fides.api.models.privacy_request.request_task import (
    AsyncTaskType,
    RequestTask,
    RequestTaskSubRequest,
)
from fides.api.models.worker_task import ExecutionLogStatus
from fides.api.schemas.policy import ActionType
from fides.api.schemas.saas.async_polling_configuration import (
    AsyncPollingConfiguration,
    PollingResult,
    PollingResultType,
)
from fides.api.schemas.saas.saas_config import ReadSaaSRequest, SaaSRequest
from fides.api.schemas.saas.shared_schemas import SaaSRequestParams
from fides.api.service.async_dsr.handlers.polling_attachment_handler import (
    PollingAttachmentHandler,
)
from fides.api.service.async_dsr.handlers.polling_request_handler import (
    PollingRequestHandler,
)
from fides.api.service.async_dsr.handlers.polling_response_handler import (
    PollingResponseProcessor,
)
from fides.api.service.async_dsr.handlers.polling_sub_request_handler import (
    PollingSubRequestHandler,
)
from fides.api.service.async_dsr.strategies.async_dsr_strategy import AsyncDSRStrategy
from fides.api.service.async_dsr.utils import AsyncPhase, get_async_phase
from fides.api.service.connectors.saas.authenticated_client import AuthenticatedClient
from fides.api.service.saas_request.saas_request_override_factory import (
    SaaSRequestOverrideFactory,
    SaaSRequestType,
)
from fides.api.util.collection_util import Row
from fides.api.util.saas_util import map_param_values
from fides.config import CONFIG

if TYPE_CHECKING:
    from fides.api.models.privacy_request.privacy_request import PrivacyRequest
    from fides.api.service.connectors.query_configs.saas_query_config import (
        SaaSQueryConfig,
    )


class AsyncPollingStrategy(AsyncDSRStrategy):
    """
    Enhanced strategy for polling async DSR requests.
    Works for both access and erasure operations with internal phase-based organization.
    """

    type = AsyncTaskType.polling
    configuration_model = AsyncPollingConfiguration

    def __init__(self, session: Session, configuration: AsyncPollingConfiguration):
        self.session = session
        self.status_request = configuration.status_request
        self.result_request = configuration.result_request

    def async_retrieve_data(
        self,
        client: AuthenticatedClient,
        request_task_id: str,
        query_config: "SaaSQueryConfig",
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """
        Execute async retrieve data with internal phase routing.
        """
        request_task = self._get_request_task(request_task_id)
        async_phase = get_async_phase(request_task, query_config)

        if async_phase == AsyncPhase.initial_async:
            return self._initial_request_access(
                client, request_task, query_config, input_data
            )

        if async_phase == AsyncPhase.polling_continuation:
            return self._polling_continuation_access(client, request_task, query_config)

        logger.warning(
            f"Unexpected async phase '{async_phase}' for polling access task {request_task.id}"
        )
        return []

    def async_mask_data(
        self,
        client: AuthenticatedClient,
        request_task_id: str,
        query_config: "SaaSQueryConfig",
        rows: List[Row],
    ) -> int:
        """
        Execute async mask data with internal phase routing.
        """
        request_task = self._get_request_task(request_task_id)
        async_phase = get_async_phase(request_task, query_config)

        if async_phase == AsyncPhase.initial_async:
            return self._initial_request_erasure(
                client, request_task, query_config, rows
            )

        if async_phase == AsyncPhase.polling_continuation:
            return self._polling_continuation_erasure(
                client, request_task, query_config
            )

        logger.warning(
            f"Unexpected async phase '{async_phase}' for polling erasure task {request_task.id}"
        )
        return 0

    # Private helper methods

    def _initial_request_access(
        self,
        client: AuthenticatedClient,
        request_task: RequestTask,
        query_config: "SaaSQueryConfig",
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """Handle initial setup for access polling requests."""
        logger.info(f"Initial polling request for access task {request_task.id}")

        policy = request_task.privacy_request.policy

        async_requests_to_process = [
            req
            for req in query_config.get_read_requests_by_identity()
            if req.async_config
        ]

        if not async_requests_to_process:
            logger.warning(
                f"No async-configured read requests found for task {request_task.id}"
            )
            return []

        request_task.async_type = AsyncTaskType.polling
        self.session.add(request_task)
        self.session.commit()

        for read_request in async_requests_to_process:
            logger.info(
                f"Creating initial polling sub-requests for task {request_task.id}"
            )

            self._handle_polling_initial_request(
                request_task,
                query_config,
                read_request,
                input_data,
                policy,
                client,
            )

        self.session.refresh(request_task)
        raise AwaitingAsyncProcessing(
            f"Waiting for next scheduled check of {request_task.dataset_name} access results."
        )

    def _initial_request_erasure(
        self,
        client: AuthenticatedClient,
        request_task: RequestTask,
        query_config: "SaaSQueryConfig",
        rows: List[Row],
    ) -> int:
        """Handle initial setup for erasure polling requests."""
        logger.info(f"Initial polling request for erasure task {request_task.id}")

        privacy_request = request_task.privacy_request
        policy = privacy_request.policy

        all_requests = []
        masking_request = query_config.get_masking_request()
        if masking_request:
            all_requests.append(masking_request)

        # Set async type once for the task
        request_task.async_type = AsyncTaskType.polling
        self.session.add(request_task)
        self.session.commit()

        for request in all_requests:
            if not (request.async_config and request_task.id):
                continue

            if request.path:
                logger.info(
                    f"Executing initial masking request for polling task {request_task.id}"
                )
                self._handle_polling_initial_erasure_request(
                    request_task,
                    query_config,
                    request,
                    rows,
                    policy,
                    privacy_request,
                    client,
                )

        # After processing all requests, raise AwaitingAsyncProcessing (like access flow)
        # But only if we actually created any sub-requests
        self.session.refresh(request_task)
        if request_task.sub_requests:
            raise AwaitingAsyncProcessing(
                f"Waiting for next scheduled check of {request_task.dataset_name} erasure results."
            )
        # If no sub-requests were created (no rows to process), the task is already complete
        return 0

    def _polling_continuation_access(
        self,
        client: AuthenticatedClient,
        request_task: RequestTask,
        query_config: "SaaSQueryConfig",
    ) -> List[Row]:
        """Handle polling continuation for access requests."""
        logger.info(f"Continuing polling for access task {request_task.id}")

        polling_complete = self._execute_polling_requests(
            client, request_task, query_config
        )

        if not polling_complete:
            raise AwaitingAsyncProcessing(
                f"Waiting for next scheduled check of {request_task.dataset_name} access results."
            )

        # Aggregate all sub-request results, merging attachments
        aggregated_results: List[Row] = []
        merged_attachments = []

        for sub_request in request_task.sub_requests:
            if sub_request.access_data:
                for row in sub_request.access_data:
                    # Check if this row has retrieved_attachments
                    if "retrieved_attachments" in row:
                        # Collect attachments for merging
                        merged_attachments.extend(row["retrieved_attachments"])
                        # Remove retrieved_attachments from the row to avoid duplication
                        row_without_attachments = {
                            k: v for k, v in row.items() if k != "retrieved_attachments"
                        }
                        # Only add the row if it's not empty or if it's the first row
                        if row_without_attachments or not aggregated_results:
                            aggregated_results.append(row_without_attachments)
                    else:
                        # Regular row without attachments
                        aggregated_results.append(row)

        # If we have merged attachments, add them to the first aggregated result
        if merged_attachments and aggregated_results:
            aggregated_results[0]["retrieved_attachments"] = merged_attachments

        return aggregated_results

    def _polling_continuation_erasure(
        self,
        client: AuthenticatedClient,
        request_task: RequestTask,
        query_config: "SaaSQueryConfig",
    ) -> int:
        """Handle polling continuation for erasure requests."""
        logger.info(f"Continuing polling for erasure task {request_task.id}")

        polling_complete = self._execute_polling_requests(
            client, request_task, query_config
        )

        if not polling_complete:
            raise AwaitingAsyncProcessing(
                f"Waiting for next scheduled check of {request_task.dataset_name} erasure results."
            )

        # Aggregate rows_masked from all sub-requests
        total_rows_masked = sum(
            sub_request.rows_masked or 0 for sub_request in request_task.sub_requests
        )
        return total_rows_masked

    def _handle_polling_initial_request(
        self,
        request_task: RequestTask,
        query_config: "SaaSQueryConfig",
        read_request: ReadSaaSRequest,
        input_data: Dict[str, Any],
        policy: Policy,
        client: "AuthenticatedClient",
    ) -> None:
        """Handles the setup for asynchronous initial requests."""
        query_config.action = "Polling - start"
        prepared_requests: List[Tuple[SaaSRequestParams, Dict[str, Any]]] = (
            query_config.generate_requests(input_data, policy, read_request)
        )
        logger.info(f"Prepared requests: {len(prepared_requests)}")

        for next_request, param_value_map in prepared_requests:
            response = client.send(next_request, read_request.ignore_errors)

            if not response.ok:
                raise FidesopsException(
                    f"Initial async request failed with status code {response.status_code}: {response.text}"
                )

            try:
                response_data = response.json()
                correlation_id = pydash.get(
                    response_data, read_request.correlation_id_path
                )
                if not correlation_id:
                    raise FidesopsException(
                        f"Could not extract correlation ID from response using path: {read_request.correlation_id_path}"
                    )
            except ValueError as exc:
                raise FidesopsException(
                    f"Invalid JSON response from initial request: {exc}"
                )

            param_value_map["correlation_id"] = str(correlation_id)
            PollingSubRequestHandler.create_sub_request(
                self.session, request_task, param_value_map
            )

    def _handle_polling_initial_erasure_request(
        self,
        request_task: RequestTask,
        query_config: "SaaSQueryConfig",
        request: SaaSRequest,
        rows: List[Row],
        policy: Policy,
        privacy_request: "PrivacyRequest",
        client: "AuthenticatedClient",
    ) -> None:
        """Handles the setup for asynchronous initial erasure requests."""
        logger.info(
            f"Processing {len(rows)} rows for erasure request in task {request_task.id}"
        )

        # Create sub-requests for erasure operations (similar to access operations)
        for row in rows or [{}]:
            try:
                # Generate parameter values first (like access requests)
                param_value_map = query_config.generate_update_param_values(
                    row, policy, privacy_request, request
                )

                # Generate the update request using the param_values
                prepared_request = query_config.generate_update_stmt(
                    row, policy, privacy_request
                )
                response = client.send(prepared_request, request.ignore_errors)

                if not response.ok:
                    raise FidesopsException(
                        f"Initial erasure request failed with status code {response.status_code}: {response.text}"
                    )

                # Extract correlation ID from response (required, like access requests)
                try:
                    response_data = response.json()
                    correlation_id = pydash.get(
                        response_data, request.correlation_id_path
                    )
                    if not correlation_id:
                        raise FidesopsException(
                            f"Could not extract correlation ID from response using path: {request.correlation_id_path}"
                        )
                except ValueError as exc:
                    raise FidesopsException(
                        f"Invalid JSON response from initial erasure request: {exc}"
                    )

                # Add correlation_id to the existing param_value_map (like access requests)
                param_value_map["correlation_id"] = str(correlation_id)
                PollingSubRequestHandler.create_sub_request(
                    self.session, request_task, param_value_map
                )

            except ValueError as exc:
                if request.skip_missing_param_values:
                    logger.debug("Skipping optional masking request: {}", exc)
                    continue
                raise exc

    def _get_requests_for_action(
        self, polling_task: RequestTask, query_config: "SaaSQueryConfig"
    ) -> List[ReadSaaSRequest]:
        """
        Get the appropriate requests based on the action type.

        Args:
            polling_task: The polling task to get requests for
            query_config: The SaaS query configuration

        Returns:
            List of ReadSaaSRequest objects for the given action type

        Raises:
            PrivacyRequestError: If action type is unsupported or masking request not found
        """
        # Validate result_request is provided for access operations
        if polling_task.action_type == ActionType.access and not self.result_request:
            raise PrivacyRequestError(
                f"result_request is required for access operations but was not provided in polling configuration for task {polling_task.id}"
            )

        if polling_task.action_type == ActionType.access:
            return list(query_config.get_read_requests_by_identity())
        if polling_task.action_type == ActionType.erasure:
            masking_request = query_config.get_masking_request()
            if not masking_request:
                raise PrivacyRequestError(
                    f"No masking request found for erasure task {polling_task.id}"
                )
            return [cast(ReadSaaSRequest, masking_request)]
        raise PrivacyRequestError(
            f"Unsupported action type: {polling_task.action_type}"
        )

    def _check_sub_request_status(
        self, client: AuthenticatedClient, param_values: Dict[str, Any]
    ) -> bool:
        """
        Check the status of a sub-request using either override function or HTTP request.

        Args:
            client: The authenticated client
            param_values: The parameter values for the request

        Returns:
            bool: True if the request is complete, False if still in progress

        Raises:
            PrivacyRequestError: If status_path is required but not provided
        """
        # Check for status override vs standard HTTP request
        if self.status_request.request_override:
            # Handle status override function directly
            override_function = SaaSRequestOverrideFactory.get_override(
                self.status_request.request_override,
                SaaSRequestType.POLLING_STATUS,
            )

            # Override functions return boolean status directly
            return cast(
                bool,
                override_function(
                    client=client,
                    param_values=param_values,
                    request_config=self.status_request,
                    secrets=client.configuration.secrets,
                ),
            )

        # Standard HTTP status request - create handler only when needed
        polling_handler = PollingRequestHandler(
            self.status_request, self.result_request
        )
        response = polling_handler.get_status_response(client, param_values)

        # Process status response
        status_path = self.status_request.status_path

        if status_path is None:
            raise PrivacyRequestError(
                "status_path is required when request_override is not provided"
            )

        return PollingResponseProcessor.evaluate_status_response(
            response,
            status_path,
            self.status_request.status_completed_value,
        )

    def _process_completed_sub_request(
        self,
        client: AuthenticatedClient,
        param_values: Dict[str, Any],
        sub_request: RequestTaskSubRequest,
        polling_task: RequestTask,
    ) -> None:
        """
        Process a completed sub-request by getting results and handling them appropriately.

        Args:
            client: The authenticated client
            param_values: The parameter values for the request
            sub_request: The completed sub-request
            polling_task: The parent polling task

        Raises:
            PrivacyRequestError: If polling result is not the expected type
        """
        # Handle erasure operations differently - they don't need result_request
        if polling_task.action_type == ActionType.erasure:
            # For erasure operations, just mark as complete and increment counter
            sub_request.rows_masked = 1
            sub_request.update_status(self.session, ExecutionLogStatus.complete.value)
            logger.info(
                f"Sub-request {sub_request.id} for {polling_task.action_type} task {polling_task.id} completed"
            )
            return

        # For access operations, we need result_request to get the data
        if not self.result_request:
            raise PrivacyRequestError(
                f"result_request is required for processing completed sub-request {sub_request.id} but was not provided"
            )

        if self.result_request.request_override:
            # Handle override function directly
            override_function = SaaSRequestOverrideFactory.get_override(
                self.result_request.request_override,
                SaaSRequestType.POLLING_RESULT,
            )

            polling_result = override_function(
                client=client,
                param_values=param_values,
                request_config=self.result_request,
                secrets=client.configuration.secrets,
            )
        else:
            # Standard HTTP request processing
            polling_handler = PollingRequestHandler(
                self.status_request, self.result_request
            )
            response = polling_handler.get_result_response(client, param_values)

            # We need to reconstruct the request path for processing
            prepared_result_request = map_param_values(
                action="result",
                context="polling request",
                current_request=self.result_request,
                param_values=param_values,
            )

            polling_result = PollingResponseProcessor.process_result_response(
                prepared_result_request.path,
                response,
                self.result_request.result_path,
            )

        # Checks if we have a polling result, response could be empty in case there was no data to access
        if polling_result:
            # Ensure we have the expected polling result type
            if not isinstance(polling_result, PollingResult):
                raise PrivacyRequestError(
                    "Polling result must be PollingResult instance"
                )

            # Store results on the sub-request
            self._store_sub_request_result(polling_result, sub_request, polling_task)
        else:
            logger.info(
                f"No result response for sub-request {sub_request.id} for task {polling_task.id}"
            )
        # Mark as complete using existing method
        sub_request.update_status(self.session, ExecutionLogStatus.complete.value)

        logger.info(
            f"Sub-request {sub_request.id} for task {polling_task.id} completed successfully"
        )

    def _process_sub_requests_for_request(
        self,
        client: AuthenticatedClient,
        request: ReadSaaSRequest,
        polling_task: RequestTask,
    ) -> None:
        """
        Process all sub-requests for a given request.

        Args:
            client: The authenticated client
            request: The SaaS request being processed
            polling_task: The parent polling task
        """
        sub_requests: List[RequestTaskSubRequest] = polling_task.sub_requests

        for sub_request in sub_requests:
            # Skip already completed sub-requests
            if sub_request.status == ExecutionLogStatus.complete.value:
                continue

            param_values = sub_request.param_values

            try:
                # Check status of the sub-request
                status = self._check_sub_request_status(client, param_values)

                if status:
                    self._process_completed_sub_request(
                        client, param_values, sub_request, polling_task
                    )
                else:
                    logger.debug(
                        f"Sub-request {sub_request.id} for task {polling_task.id} still not ready"
                    )

            except Exception as exc:
                logger.error(
                    f"Error processing sub-request {sub_request.id} for task {polling_task.id}: {exc}"
                )
                sub_request.update_status(self.session, ExecutionLogStatus.error.value)
                raise exc

    def _execute_polling_requests(
        self,
        client: AuthenticatedClient,
        polling_task: RequestTask,
        query_config: "SaaSQueryConfig",
    ) -> bool:
        """
        Internal polling execution orchestrator with proper error handling and timeout management.

        Stores results on individual sub-requests and only aggregates when ALL are successful.
        Implements timeout checking and raises exceptions for errors and timeouts.

        Returns:
            bool: True if all polling is complete (success or failure), False if still in progress
        """
        # Check for timeout before processing requests
        timeout_days = CONFIG.execution.async_polling_request_timeout_days
        PollingSubRequestHandler.check_timeout(polling_task, timeout_days)

        # Get appropriate requests based on action type
        requests = self._get_requests_for_action(polling_task, query_config)

        # Process each request and its sub-requests
        for request in requests:
            if request.async_config:
                self._process_sub_requests_for_request(client, request, polling_task)

        # Check final status and return completion status
        return PollingSubRequestHandler.check_completion(polling_task)

    def _store_sub_request_result(
        self,
        polling_result: PollingResult,
        sub_request: RequestTaskSubRequest,
        polling_task: RequestTask,
    ) -> None:
        """Store result data on the individual sub-request."""
        if polling_result.result_type == PollingResultType.rows:
            # Store rows directly on the sub-request (data is always a list for rows)
            sub_request.access_data = polling_result.data
            sub_request.save(self.session)

        elif polling_result.result_type == PollingResultType.attachment:
            try:
                attachment_bytes = PollingAttachmentHandler.ensure_attachment_bytes(
                    polling_result.data
                )
                attachment_id = PollingAttachmentHandler.store_attachment(
                    self.session,
                    polling_task,
                    attachment_data=attachment_bytes,
                    filename=polling_result.metadata.get(
                        "filename", f"attachment_{str(uuid4())[:8]}"
                    ),
                )

                # Store attachment metadata on sub-request
                attachment_metadata: List[Row] = [{"retrieved_attachments": []}]
                PollingAttachmentHandler.add_metadata_to_rows(
                    self.session, attachment_id, attachment_metadata
                )
                sub_request.access_data = attachment_metadata
                sub_request.save(self.session)
            except Exception as exc:
                raise PrivacyRequestError(f"Attachment storage failed: {exc}")
        else:
            raise PrivacyRequestError(
                f"Unsupported result type: {polling_result.result_type}"
            )



from enum import Enum
from typing import Optional, Union

from pydantic import BaseModel, Field, field_validator


class PrivacyRequestTopLevelFields(Enum):
    """Top level fields for privacy request."""

    privacy_request = "privacy_request"
    policy = "privacy_request.policy"
    identity = "privacy_request.identity"


class PrivacyRequestPolicyConvenienceFields(Enum):
    """Convenience fields for privacy request policy."""

    rule_action_types = "rule_action_types"
    has_access_rule = "has_access_rule"
    has_erasure_rule = "has_erasure_rule"
    has_consent_rule = "has_consent_rule"
    has_update_rule = "has_update_rule"
    rule_count = "rule_count"
    rule_names = "rule_names"
    has_storage_destination = "has_storage_destination"


class PrivacyRequestLocationConvenienceFields(Enum):
    """Convenience fields for privacy request location."""

    location_country = "location_country"
    location_groups = "location_groups"
    location_regulations = "location_regulations"


class PrivacyRequestConvenienceFields(Enum):
    """Convenience fields for privacy request."""

    # Policy convenience fields
    rule_action_types = f"{PrivacyRequestTopLevelFields.policy.value}.{PrivacyRequestPolicyConvenienceFields.rule_action_types.value}"
    has_access_rule = f"{PrivacyRequestTopLevelFields.policy.value}.{PrivacyRequestPolicyConvenienceFields.has_access_rule.value}"
    has_erasure_rule = f"{PrivacyRequestTopLevelFields.policy.value}.{PrivacyRequestPolicyConvenienceFields.has_erasure_rule.value}"
    has_consent_rule = f"{PrivacyRequestTopLevelFields.policy.value}.{PrivacyRequestPolicyConvenienceFields.has_consent_rule.value}"
    has_update_rule = f"{PrivacyRequestTopLevelFields.policy.value}.{PrivacyRequestPolicyConvenienceFields.has_update_rule.value}"
    rule_count = f"{PrivacyRequestTopLevelFields.policy.value}.{PrivacyRequestPolicyConvenienceFields.rule_count.value}"
    rule_names = f"{PrivacyRequestTopLevelFields.policy.value}.{PrivacyRequestPolicyConvenienceFields.rule_names.value}"
    has_storage_destination = f"{PrivacyRequestTopLevelFields.policy.value}.{PrivacyRequestPolicyConvenienceFields.has_storage_destination.value}"
    # Location hierarchy convenience fields
    location_country = f"{PrivacyRequestTopLevelFields.privacy_request.value}.{PrivacyRequestLocationConvenienceFields.location_country.value}"
    location_groups = f"{PrivacyRequestTopLevelFields.privacy_request.value}.{PrivacyRequestLocationConvenienceFields.location_groups.value}"
    location_regulations = f"{PrivacyRequestTopLevelFields.privacy_request.value}.{PrivacyRequestLocationConvenienceFields.location_regulations.value}"


class PrivacyRequestFields(Enum):
    """Fields for privacy request."""

    created_at = f"{PrivacyRequestTopLevelFields.privacy_request.value}.created_at"
    due_date = f"{PrivacyRequestTopLevelFields.privacy_request.value}.due_date"
    identity_verified_at = (
        f"{PrivacyRequestTopLevelFields.privacy_request.value}.identity_verified_at"
    )
    location = f"{PrivacyRequestTopLevelFields.privacy_request.value}.location"
    origin = f"{PrivacyRequestTopLevelFields.privacy_request.value}.origin"
    requested_at = f"{PrivacyRequestTopLevelFields.privacy_request.value}.requested_at"
    source = f"{PrivacyRequestTopLevelFields.privacy_request.value}.source"
    submitted_by = f"{PrivacyRequestTopLevelFields.privacy_request.value}.submitted_by"


class PolicyFields(Enum):
    """Fields for policy."""

    id = "privacy_request.policy.id"
    name = f"{PrivacyRequestTopLevelFields.policy.value}.name"
    key = f"{PrivacyRequestTopLevelFields.policy.value}.key"
    description = f"{PrivacyRequestTopLevelFields.policy.value}.description"
    execution_timeframe = (
        f"{PrivacyRequestTopLevelFields.policy.value}.execution_timeframe"
    )
    rules = f"{PrivacyRequestTopLevelFields.policy.value}.rules"


class IdentityFields(Enum):
    """Fields for identity."""

    email = f"{PrivacyRequestTopLevelFields.identity.value}.email"
    phone_number = f"{PrivacyRequestTopLevelFields.identity.value}.phone_number"
    external_id = f"{PrivacyRequestTopLevelFields.identity.value}.external_id"
    fides_user_device_id = (
        f"{PrivacyRequestTopLevelFields.identity.value}.fides_user_device_id"
    )
    ljt_readerID = f"{PrivacyRequestTopLevelFields.identity.value}.ljt_readerID"
    ga_client_id = f"{PrivacyRequestTopLevelFields.identity.value}.ga_client_id"


class ConditionalDependencyFieldType(Enum):
    """Type of the field for conditional dependencies."""

    string = "string"
    boolean = "boolean"
    array = "array"
    integer = "integer"
    float = "float"
    date = "date"


class CustomFieldPathPrefix(Enum):
    """Valid path prefixes for custom fields in privacy request conditional dependencies.

    Custom privacy request fields are dynamically created based on privacy center
    configuration and are stored on the privacy request itself (not the policy).
    """

    custom_privacy_request_fields = "privacy_request.custom_privacy_request_fields"

    @classmethod
    def is_valid_custom_field_path(cls, field_path: str) -> bool:
        """Check if a field path is a valid custom field path.

        Args:
            field_path: The field path to validate

        Returns:
            True if the path starts with a valid custom field prefix, False otherwise
        """
        return any(field_path.startswith(prefix.value + ".") for prefix in cls)

    @classmethod
    def get_custom_field_name(cls, field_path: str) -> Optional[str]:
        """Extract the custom field name from a valid custom field path.

        Args:
            field_path: The field path to extract the field name from

        Returns:
            The custom field name if the path is valid, None otherwise
        """
        for prefix in cls:
            if field_path.startswith(prefix.value + "."):
                return field_path[len(prefix.value) + 1 :]
        return None


# Union type for all valid field paths
# Pydantic will validate against enum values for standard fields
# Custom field paths (dynamically created, validated at runtime) follow the pattern:
# privacy_request.custom_privacy_request_fields.{field_name}
ConditionalDependencyFieldPath = Union[
    PrivacyRequestFields,
    PolicyFields,
    IdentityFields,
    PrivacyRequestConvenienceFields,
    str,  # Custom field paths (must match prefix pattern, validated below)
]


class ConditionalDependencyFieldInfo(BaseModel):
    """Information about a field available for conditional dependencies."""

    field_path: ConditionalDependencyFieldPath = Field(
        ...,
        description="Path to the field in the PrivacyRequest. Must be one of the standard field paths or a valid custom field path.",
    )
    field_type: ConditionalDependencyFieldType = Field(
        ..., description="Type of the field from ConditionalDependencyFieldType"
    )
    description: Optional[str] = Field(None, description="Description of the field")
    is_convenience_field: bool = Field(
        False,
        description="Whether this is a convenience field (derived from other fields)",
    )

    @field_validator("field_path")
    @classmethod
    def validate_field_path(cls, value: Union[Enum, str]) -> Union[Enum, str]:
        """Validate that field_path is either an enum value or a valid custom field path.

        Args:
            value: The field path to validate (can be an enum member or string)

        Returns:
            The validated field path

        Raises:
            ValueError: If the field path is not valid
        """
        # If it's already an enum member, it's valid
        if isinstance(value, Enum):
            return value

        # If it's a string, check if it matches an enum value or is a valid custom field path
        if isinstance(value, str):
            # Check if it matches any enum value
            all_enum_values = (
                {field.value for field in PrivacyRequestFields}
                | {field.value for field in PolicyFields}
                | {field.value for field in IdentityFields}
                | {field.value for field in PrivacyRequestConvenienceFields}
            )

            if value in all_enum_values:
                return value

            # If it doesn't match an enum value, it must be a valid custom field path
            if not CustomFieldPathPrefix.is_valid_custom_field_path(value):
                raise ValueError(
                    f"Invalid field_path '{value}'. Must be one of the standard field paths "
                    f"or a valid custom field path starting with one of: "
                    f"{', '.join(prefix.value for prefix in CustomFieldPathPrefix)}"
                )
            return value

        return value



from __future__ import annotations

from datetime import datetime
from enum import Enum
from re import match
from typing import Any, Dict, Iterable, List, Optional, Set, Type

from loguru import logger
from sqlalchemy import (
    ARRAY,
    Boolean,
    CheckConstraint,
    Column,
    DateTime,
    ForeignKey,
    Index,
    String,
    UniqueConstraint,
    func,
    text,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.future import select
from sqlalchemy.orm import RelationshipProperty, Session, relationship, validates
from sqlalchemy.orm.query import Query

from fides.api.db.base_class import Base, FidesBase
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.detection_discovery.staged_resource_error import (
    StagedResourceError,
)
from fides.api.models.sql_models import System  # type: ignore[attr-defined]


class DiffStatus(Enum):
    ADDITION = "addition"
    REMOVAL = "removal"
    CLASSIFICATION_ADDITION = "classification_addition"
    CLASSIFICATION_UPDATE = "classification_update"
    MONITORED = "monitored"
    MUTED = "muted"


class MonitorFrequency(Enum):
    """
    Enum representing monitor frequency. Not used in DB but needed for translating to API schema
    """

    DAILY = "Daily"
    WEEKLY = "Weekly"
    MONTHLY = "Monthly"
    QUARTERLY = "Quarterly"
    YEARLY = "Yearly"
    NOT_SCHEDULED = "Not scheduled"


# pattern for a string of 4 comma-separated integers,
# used to represent the months of the year that the monitor will run
# on quarterly basis, in cron format
QUARTERLY_MONTH_PATTERN = r"^\d+,\d+,\d+,\d+$"


class StagedResourceType(str, Enum):
    """
    Enum representing the type of staged resource.
    The resource_type column is a string in the DB, this is just for
    application-level use.
    """

    # Note: If you add a new resource type, make sure to update either
    # get_datastore_resource_types or get_website_monitor_resource_types

    # Datastore staged resources
    DATABASE = "Database"
    SCHEMA = "Schema"
    TABLE = "Table"
    FIELD = "Field"
    ENDPOINT = "Endpoint"
    # Website monitor staged resources
    COOKIE = "Cookie"
    BROWSER_REQUEST = "Browser request"
    IMAGE_BROWSER_REQUEST = "Image"
    IFRAME_BROWSER_REQUEST = "iFrame"
    JAVASCRIPT_BROWSER_REQUEST = "Javascript tag"

    @staticmethod
    def get_datastore_resource_types() -> List["StagedResourceType"]:
        return [
            StagedResourceType.DATABASE,
            StagedResourceType.SCHEMA,
            StagedResourceType.TABLE,
            StagedResourceType.FIELD,
            StagedResourceType.ENDPOINT,
        ]

    @staticmethod
    def get_website_monitor_resource_types() -> List["StagedResourceType"]:
        return [
            StagedResourceType.COOKIE,
            StagedResourceType.BROWSER_REQUEST,
            StagedResourceType.IMAGE_BROWSER_REQUEST,
            StagedResourceType.IFRAME_BROWSER_REQUEST,
            StagedResourceType.JAVASCRIPT_BROWSER_REQUEST,
        ]

    def is_datastore_resource(self) -> bool:
        return self in self.get_datastore_resource_types()

    def is_website_monitor_resource(self) -> bool:
        return self in self.get_website_monitor_resource_types()


class SharedMonitorConfig(Base, FidesBase):
    """SQL model for shareable monitor configurations"""

    @declared_attr
    def __tablename__(self) -> str:
        return "shared_monitor_config"

    # Basic info
    name = Column(String, nullable=False)
    key = Column(String, unique=True, nullable=False)
    description = Column(String, nullable=True)

    # Classification parameters (including regex patterns)
    classify_params = Column(
        MutableDict.as_mutable(JSONB),
        index=False,
        unique=False,
        nullable=False,
        server_default="{}",
        default=dict,
    )


class MonitorConfig(Base):
    """
    Monitor configuration used for data detection and discovery.

    Each monitor configuration references `ConnectionConfig`, which provide it with underlying
    configuration details used in connecting to the external data store.
    """

    name = Column(String, nullable=True)
    key = Column(String, index=True, unique=True, nullable=False)
    connection_config_id = Column(
        String,
        ForeignKey(ConnectionConfig.id_field_path),
        nullable=False,
        index=True,
    )
    databases = Column(
        ARRAY(String),
        index=False,
        unique=False,
        nullable=False,
        server_default="{}",
        default=dict,
    )  # the databases to which the monitor is scoped
    excluded_databases = Column(
        ARRAY(String),
        index=False,
        unique=False,
        nullable=False,
        server_default="{}",
        default=dict,
    )  # the databases to which the monitor is not scoped
    monitor_execution_trigger = Column(
        MutableDict.as_mutable(JSONB),
        index=False,
        unique=False,
        nullable=True,
    )  # stores the cron-based kwargs for scheduling the monitor execution.
    # see https://apscheduler.readthedocs.io/en/3.x/modules/triggers/cron.html

    # We use _classify_params for the actual column to prevent direct access
    _classify_params = Column(
        "classify_params",
        MutableDict.as_mutable(JSONB),
        index=False,
        unique=False,
        nullable=True,
    )  # parameters that the monitor will use for classification execution

    datasource_params = Column(
        MutableDict.as_mutable(JSONB),
        index=False,
        unique=False,
        nullable=True,
    )  # monitor parameters that are specific per datasource
    # these are held as an untyped JSON dict (in the DB) to stay flexible

    last_monitored = Column(
        DateTime(timezone=True),
        nullable=True,
    )  # when the monitor was last executed

    enabled = Column(
        Boolean,
        default=True,
        nullable=False,
        server_default="t",
    )

    # TODO: many-to-many link to users assigned as data stewards; likely will need a join-table

    connection_config = relationship(ConnectionConfig)

    executions = relationship(
        "MonitorExecution",
        cascade="all, delete-orphan",
        backref="monitor_config",
        primaryjoin="MonitorExecution.monitor_config_key == foreign(MonitorConfig.key)",
        single_parent=True,
    )

    shared_config_id = Column(
        String,
        ForeignKey(SharedMonitorConfig.id_field_path, ondelete="RESTRICT"),
        nullable=True,
        index=True,
    )

    shared_config = relationship(SharedMonitorConfig)

    __table_args__ = (
        CheckConstraint(  # type: ignore
            "key NOT LIKE '%.%'", name="ck_monitorconfig_key_no_dots"
        ),
    )

    @property
    def classify_params(self) -> dict:
        """
        Returns the merged classify parameters from both the monitor config and
        the shared config (if it exists).

        The shared config parameters take precedence over the monitor's own parameters,
        but only for values that are not falsy (None, empty, etc.).
        """
        # Start with an empty dict
        merged_params = {}

        # Add this monitor's params if available
        if self._classify_params:
            merged_params.update(self._classify_params)

        # Add/override with shared config params if available
        if self.shared_config and self.shared_config.classify_params:
            # Only update with non-falsy values from shared config
            for key, value in self.shared_config.classify_params.items():
                if value:  # Only override if the value is not falsy
                    merged_params[key] = value

        return merged_params

    @classify_params.setter
    def classify_params(self, value: Dict[str, Any]) -> None:
        """Setter for the classify_params to maintain compatibility with existing code"""
        self._classify_params = value

    @property
    def connection_config_key(self) -> str:
        """Derives the `connection_config_key`"""
        return self.connection_config.key

    @property
    def execution_start_date(self) -> Optional[datetime]:
        """Derives the `execution_start_date`"""
        if (
            not self.monitor_execution_trigger
            or self.monitor_execution_trigger.get("start_date", None) is None
        ):
            return None
        return self.monitor_execution_trigger.get("start_date")

    @property
    def execution_frequency(self) -> Optional[MonitorFrequency]:
        """Derives the `execution_frequency`"""
        if (
            not self.monitor_execution_trigger
            or self.monitor_execution_trigger.get("hour", None) is None
        ):
            return MonitorFrequency.NOT_SCHEDULED
        month_trigger = self.monitor_execution_trigger.get("month", None)
        if month_trigger is not None:
            if isinstance(month_trigger, str) and match(
                QUARTERLY_MONTH_PATTERN, month_trigger
            ):
                return MonitorFrequency.QUARTERLY
            return MonitorFrequency.YEARLY
        if self.monitor_execution_trigger.get("day", None) is not None:
            return MonitorFrequency.MONTHLY
        if self.monitor_execution_trigger.get("day_of_week", None) is not None:
            return MonitorFrequency.WEEKLY
        return MonitorFrequency.DAILY

    def update(self, db: Session, *, data: dict[str, Any]) -> FidesBase:
        """
        Override the base class `update` to validate database include/exclude
        and derive the `execution_trigger` dict field
        """
        MonitorConfig.database_include_exclude_list_is_valid(data)
        MonitorConfig.derive_execution_trigger_dict(data)
        return super().update(db=db, data=data)

    @classmethod
    def database_include_exclude_list_is_valid(cls, data: Dict[str, Any]) -> None:
        """Check that both include and exclude have not both been set"""
        include = data.get("databases", [])
        exclude = data.get("excluded_databases", [])
        if include and exclude:
            raise ValueError(
                "Both `databases` and `excluded_databases` cannot be set at the same time."
            )

    @classmethod
    def create(
        cls: Type[MonitorConfig],
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = True,
    ) -> MonitorConfig:
        """
        Override the base class `create` to validate database include/exclude
        and derive the `execution_trigger` dict field
        """
        MonitorConfig.database_include_exclude_list_is_valid(data)
        MonitorConfig.derive_execution_trigger_dict(data)
        return super().create(db=db, data=data, check_name=check_name)

    @staticmethod
    def derive_execution_trigger_dict(data: Dict[str, Any]) -> None:
        """
        Determines the execution trigger (cron) dict based on the
        corresponding schema properties provided in the `data` dict.

        The `data` dict is updated in place with the execution trigger dict
        placed in the `monitor_execution_trigger` key, if applicable, and with the
        `execution_frequency` and `execution_start_date` keys removed.

        The `execution_start_date` is inferred as the basis for the day and time
        for repeated monitor execution, and the frequency of execution is based
        on the `execution_frequency` field.

        For example, an `execution_start_date` of "2024-05-14 12:00:00+00:00":
        - with an `execution_frequency` of "daily", it will result in daily
        execution at 12:00:00+00:00;
        - with an `execution_frequency` of "weekly", it will result in weekly
        execution at 12:00:00+00:00 on every Tuesday, since 2024-05-14 is
        a Tuesday.
        - with an `execution_frequency` of "monthly", it will result in monthly
        execution at 12:00:00+00:00 on the 14th day of every month.
        - with an `execution_frequency` of "quarterly", it will result in quarterly
        execution at 12:00:00+00:00 on the 14th day of the first month of each quarter.
        - with an `execution_frequency` of "yearly", it will result in yearly
        execution at 12:00:00+00:00 on May 14th of each year.

        See https://apscheduler.readthedocs.io/en/3.x/modules/triggers/cron.html
        for more information about the cron trigger parameters.
        """
        execution_frequency = data.pop("execution_frequency", None)
        execution_start_date = data.pop("execution_start_date", None)
        if execution_frequency == MonitorFrequency.NOT_SCHEDULED:
            data["monitor_execution_trigger"] = None
            return
        if execution_frequency and execution_start_date:
            cron_trigger_dict = {}
            cron_trigger_dict["start_date"] = execution_start_date
            cron_trigger_dict["timezone"] = execution_start_date.tzinfo
            cron_trigger_dict["hour"] = execution_start_date.hour
            cron_trigger_dict["minute"] = execution_start_date.minute
            cron_trigger_dict["second"] = execution_start_date.second
            if execution_frequency == MonitorFrequency.WEEKLY:
                cron_trigger_dict["day_of_week"] = execution_start_date.weekday()
            if execution_frequency == MonitorFrequency.MONTHLY:
                cron_trigger_dict["day"] = execution_start_date.day
            if execution_frequency == MonitorFrequency.QUARTERLY:
                cron_trigger_dict["day"] = execution_start_date.day
                # Calculate which month of the quarter (0-2) this is
                month_of_quarter = (execution_start_date.month - 1) % 3
                # Set to run in the same month of each quarter (1, 4, 7, 10 for first month)
                cron_trigger_dict["month"] = (
                    f"{1 + month_of_quarter},{4 + month_of_quarter},{7 + month_of_quarter},{10 + month_of_quarter}"
                )
            if execution_frequency == MonitorFrequency.YEARLY:
                cron_trigger_dict["day"] = execution_start_date.day
                cron_trigger_dict["month"] = execution_start_date.month
            data["monitor_execution_trigger"] = cron_trigger_dict

    @validates("key")
    def validate_key_no_dots(self, _key: str, value: str) -> str:
        if value and "." in value:
            raise ValueError('MonitorConfig.key cannot contain "." characters')
        return value


class StagedResourceAncestor(Base):
    """
    A simple junction table that is used to store the many-to-many relationship
    between staged resources and their ancestors.

    This table is used to easily lookup all ancestors of a given staged resource,
    as its indexed by both ancestor and descendant URN columns.

    Its entries should be deleted when the staged resource is deleted, via cascade.
    """

    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)

    ancestor_urn = Column(
        String,
        ForeignKey("stagedresource.urn", ondelete="CASCADE"),
        nullable=False,
    )
    descendant_urn = Column(
        String,
        ForeignKey("stagedresource.urn", ondelete="CASCADE"),
        nullable=False,
    )

    ancestor_staged_resource = relationship(
        "StagedResource",
        back_populates="ancestor_links",
        lazy="selectin",
        foreign_keys=[ancestor_urn],
    )
    descendant_staged_resource = relationship(
        "StagedResource",
        back_populates="descendant_links",
        lazy="selectin",
        foreign_keys=[descendant_urn],
    )

    __table_args__ = (
        UniqueConstraint(
            "ancestor_urn", "descendant_urn", name="uq_staged_resource_ancestor"
        ),
        Index("ix_staged_resource_ancestor_pkey", "id", unique=True),
        Index("ix_staged_resource_ancestor_ancestor", "ancestor_urn"),
        Index("ix_staged_resource_ancestor_descendant", "descendant_urn"),
    )

    @classmethod
    def create_all_staged_resource_ancestor_links(
        cls,
        db: Session,
        ancestor_links: Dict[str, Set[str]],
        batch_size: int = 10000,  # Conservative batch size
    ) -> None:
        """
        Bulk inserts all entries in the StagedResourceAncestor table
        based on the provided mapping of descendant URNs to their ancestor URN sets.

        We execute the bulk INSERT with the provided (synchronous) db session,
        but the transaction is _not_ committed, so the caller must commit the transaction
        to persist the changes.

        Uses batching to handle large datasets without hitting PostgreSQL parameter limits.

        Args:
            db: Database session
            ancestor_links: Dict mapping descendant URNs to sets of ancestor URNs
        """
        stmt_text = text(
            """
            INSERT INTO stagedresourceancestor (id, ancestor_urn, descendant_urn)
            VALUES ('srl_' || gen_random_uuid(), :ancestor_urn, :descendant_urn)
            ON CONFLICT (ancestor_urn, descendant_urn) DO NOTHING;
            """
        )

        current_batch = []

        for descendant_urn, ancestor_urns in ancestor_links.items():
            if ancestor_urns:  # Only create links if there are ancestors
                for ancestor_urn in ancestor_urns:
                    current_batch.append(
                        {"ancestor_urn": ancestor_urn, "descendant_urn": descendant_urn}
                    )

                    # Execute batch when it reaches the desired size
                    if len(current_batch) >= batch_size:
                        logger.debug(
                            f"Inserting batch of {len(current_batch)} staged resource ancestor links"
                        )
                        db.execute(stmt_text, current_batch)
                        current_batch = []

        # Execute any remaining items in the final batch
        if current_batch:
            logger.debug(
                f"Inserting batch of {len(current_batch)} staged resource ancestor links"
            )
            db.execute(stmt_text, current_batch)


class StagedResource(Base):
    """
    Base DB model that represents a staged resource, fields common to all types of staged resources
    """

    name = Column(String, nullable=True)
    urn = Column(String, index=True, unique=True, nullable=False)
    resource_type = Column(String, index=True, nullable=True)
    description = Column(String, nullable=True)
    monitor_config_id = Column(
        String,
        index=True,  # indexed because we frequently need to slice by monitor config ID
        nullable=True,
    )  # just a "soft" pointer, for now TODO: make this a FK

    # for now, this is just used for web monitor resources.
    system_id = Column(
        String,
        ForeignKey(System.id_field_path),
        nullable=True,
        index=True,
    )

    # TODO: we should be able to enable the below relationship, but it
    # confuses different functionality since 'system' means different
    # things depending on whether the resource is a datastore or web monitor
    # staged resource
    #    system = relationship(System)

    # the Compass vendor ID associated with the StagedResource.
    # only used for web monitor resources
    vendor_id = Column(
        String,
        nullable=True,
        index=True,  # indexed because we frequently need to slice by vendor ID
    )

    source_modified = Column(
        DateTime(timezone=True),
        nullable=True,
    )  # when the resource was modified in the datasource
    classifications = Column(
        ARRAY(JSONB),
        nullable=False,
        server_default="{}",
        default=dict,
    )
    user_assigned_data_categories = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=dict,
    )
    # This field is intentionally nullable to distinguish the case when no user assigned data uses
    # have been set (default, value is None) from the case in which they have been explicitly set
    # as empty (value is empty array). This allows users to remove all data uses from a StagedResource,
    # which was not possible when this field was not nullable.
    user_assigned_data_uses = Column(
        ARRAY(String),
        nullable=True,
        server_default=None,
        default=None,
    )
    user_assigned_system_id = Column(String, nullable=True, index=True)

    # pointers to child and parent URNs
    children = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=dict,
    )
    parent = Column(String, nullable=True)

    # diff-related fields
    diff_status = Column(String, nullable=True, index=True)

    errors: RelationshipProperty[List[StagedResourceError]] = relationship(
        "StagedResourceError",
        foreign_keys=[StagedResourceError.staged_resource_urn],
        primaryjoin="StagedResource.urn == StagedResourceError.staged_resource_urn",
        cascade="all, delete-orphan",
    )

    ancestor_links: RelationshipProperty[List[StagedResourceAncestor]] = relationship(
        "StagedResourceAncestor",
        back_populates="descendant_staged_resource",
        lazy="dynamic",
        cascade="all, delete",
        foreign_keys=[StagedResourceAncestor.descendant_urn],
    )

    descendant_links: RelationshipProperty[List[StagedResourceAncestor]] = relationship(
        "StagedResourceAncestor",
        back_populates="ancestor_staged_resource",
        lazy="dynamic",
        cascade="all, delete",
        foreign_keys=[StagedResourceAncestor.ancestor_urn],
    )

    def ancestors(self, db: Session) -> List[StagedResource]:
        """
        Returns the ancestors of the staged resource
        """
        # Single query to get all ancestors with their data
        query = (
            select(StagedResource)
            .join(
                StagedResourceAncestor,
                StagedResource.urn == StagedResourceAncestor.ancestor_urn,
            )
            .where(StagedResourceAncestor.descendant_urn == self.urn)
        )

        result = db.execute(query)
        return list(result.scalars().all())

    def descendants(self, db: Session) -> List[StagedResource]:
        """
        Returns the descendants of the staged resource
        """
        # Single query to get all descendants with their data
        query = (
            select(StagedResource)
            .join(
                StagedResourceAncestor,
                StagedResource.urn == StagedResourceAncestor.descendant_urn,
            )
            .where(StagedResourceAncestor.ancestor_urn == self.urn)
        )

        result = db.execute(query)
        return list(result.scalars().all())

    # placeholder for additional attributes
    meta = Column(
        MutableDict.as_mutable(JSONB),
        nullable=False,
        server_default="{}",
        default=dict,
    )

    data_uses = Column(
        ARRAY(String),
        nullable=True,
        server_default="{}",
        default=dict,
    )

    __table_args__ = (
        # Used for querying monitor aggregates
        Index(
            "ix_stagedresource_monitor_config_resource_type_consent",
            "monitor_config_id",
            "resource_type",
            text("(meta->>'consent_aggregated')"),
        ),
        # Used for querying system aggregates
        Index(
            "ix_stagedresource_system_vendor_consent",
            "system_id",
            "vendor_id",
            text("(meta->>'consent_aggregated')"),
        ),
        # GIN indices for array operations (&&, @>, <@)
        Index(
            "idx_stagedresource_classifications_gin",
            "classifications",
            postgresql_using="gin",
        ),
        Index(
            "idx_stagedresource_user_categories_gin",
            "user_assigned_data_categories",
            postgresql_using="gin",
        ),
    )

    @classmethod
    def get_urn(cls, db: Session, urn: str) -> Optional[StagedResource]:
        """Utility to retrieve the staged resource with the given URN"""
        return cls.get_by(db=db, field="urn", value=urn)

    @classmethod
    def get_urn_list(cls, db: Session, urns: Iterable[str]) -> Iterable[StagedResource]:
        """
        Utility to retrieve all staged resources with the given URNs
        """
        results = db.execute(select(StagedResource).where(StagedResource.urn.in_(urns)))
        return results.scalars().all()

    @classmethod
    async def get_urn_async(
        cls, db: AsyncSession, urn: str
    ) -> Optional[StagedResource]:
        """
        Utility to retrieve the staged resource with the given URN using an async session
        """
        results = await db.execute(
            select(StagedResource).where(StagedResource.urn == urn)
        )
        return results.scalars().first()

    @classmethod
    async def get_urn_list_async(
        cls, db: AsyncSession, urns: List[str]
    ) -> Optional[List[StagedResource]]:
        """
        Utility to retrieve the staged resource with the given URN using an async session
        """
        results = await db.execute(
            select(StagedResource).where(StagedResource.urn.in_(urns))
        )
        return results.scalars().all()

    def mark_as_addition(
        self,
        db: Session,
        parent_resource_urns: Iterable[str] = [],
    ) -> None:
        """
        Marks the resource as an addition
        """
        self.diff_status = DiffStatus.ADDITION.value


class MonitorExecution(Base):
    """
    Monitor execution record used for data detection and discovery.

    Each monitor execution references `MonitorConfig`, which provide it with underlying
    configuration details used in connecting to the external data store.
    """

    # redefined here because there's a minor, unintended discrepancy between
    # this `id` field and that of the `Base` class, which explicitly sets `index=True`.
    # TODO: we likely should _not_ be setting `index=True` on the `id`
    # attribute of the `Base` class, as `primary_key=True` already specifies a
    # primary key constraint, which will implicitly create an index for the field.
    id = Column(String(255), primary_key=True, default=FidesBase.generate_uuid)

    monitor_config_key = Column(String, nullable=False, index=True)
    status = Column(String, nullable=True)
    started = Column(
        DateTime(timezone=True),
        nullable=True,
        server_default=func.now(),
    )
    completed = Column(DateTime(timezone=True), nullable=True)
    classification_instances = Column(
        ARRAY(String),
        index=False,
        unique=False,
        nullable=False,
        default=list,
    )
    # stores additional information from monitor execution failures as an array of strings,
    # e.g. error messages, stack traces, etc.
    messages = Column(
        ARRAY(String),
        nullable=False,
        default=list,
    )
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
    )


def fetch_staged_resources_by_type_query(
    resource_type: str,
    monitor_config_ids: Optional[List[str]] = None,
    show_hidden: bool = False,
) -> Query[StagedResource]:
    """
    Fetches staged resources by type and monitor config ID. Optionally filters out muted staged resources ("hidden").
    """
    logger.info(
        f"Fetching staged resources of type {resource_type}, show_hidden={show_hidden}, monitor_config_ids={monitor_config_ids}"
    )
    query = select(StagedResource).where(StagedResource.resource_type == resource_type)

    if monitor_config_ids:
        query = query.filter(StagedResource.monitor_config_id.in_(monitor_config_ids))
    if not show_hidden:
        from sqlalchemy import or_

        query = query.filter(
            or_(
                StagedResource.diff_status != DiffStatus.MUTED.value,
                StagedResource.diff_status.is_(None),
            )
        )

    return query



from typing import TYPE_CHECKING, Any, Dict, List, Optional

from sqlalchemy import Column, DateTime, ForeignKey, Integer, String, Text
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session, relationship
from sqlalchemy.orm.attributes import flag_modified
from sqlalchemy.sql import func

from fides.api.db.base_class import Base
from fides.api.models.worker_task import ExecutionLogStatus, WorkerTask

if TYPE_CHECKING:
    from fides.api.models.digest.digest_config import DigestConfig


class DigestTaskExecution(
    WorkerTask, Base
):  # pylint: disable=too-many-instance-attributes
    """
    Model for tracking digest task execution state and progress.

    This model enables graceful resumption of digest tasks after worker
    interruptions by persisting execution state and progress information.
    """

    @declared_attr
    def __tablename__(cls) -> str:
        return "digest_task_execution"

    # Foreign key to digest config
    digest_config_id = Column(
        String,
        ForeignKey("digest_config.id", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )

    # Celery task tracking
    celery_task_id = Column(String, nullable=True, index=True)

    # Progress tracking
    total_recipients = Column(Integer, nullable=True)
    processed_recipients = Column(Integer, nullable=False, default=0)
    successful_communications = Column(Integer, nullable=False, default=0)
    failed_communications = Column(Integer, nullable=False, default=0)

    # State persistence for resumption
    execution_state = Column(JSONB, nullable=True, default={})
    processed_user_ids = Column(JSONB, nullable=True, default=[])

    # Timing information
    started_at = Column(DateTime(timezone=True), nullable=True)
    completed_at = Column(DateTime(timezone=True), nullable=True)
    last_checkpoint_at = Column(DateTime(timezone=True), nullable=True)

    # Error information
    error_message = Column(Text, nullable=True)

    # Relationships
    digest_config = relationship("DigestConfig", back_populates="executions")

    @classmethod
    def allowed_action_types(cls) -> List[str]:
        """Return allowed action types for digest task execution."""
        return ["digest_processing"]

    def mark_started(self, db: Session, celery_task_id: str) -> None:
        """Mark the execution as started."""
        self.status = ExecutionLogStatus.in_processing
        self.celery_task_id = celery_task_id
        self.started_at = func.now()
        self.save(db)

    def mark_awaiting_processing(self, db: Session) -> None:
        """Mark the execution as awaiting processing."""
        self.status = ExecutionLogStatus.awaiting_processing
        self.save(db)

    def update_progress(
        self,
        db: Session,
        processed_count: int,
        successful_count: int,
        failed_count: int,
        processed_user_ids: List[str],
        execution_state: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Update execution progress and create checkpoint."""
        self.processed_recipients = processed_count
        self.successful_communications = successful_count
        self.failed_communications = failed_count
        self.processed_user_ids = processed_user_ids
        self.last_checkpoint_at = func.now()

        # Mark JSONB fields as modified so SQLAlchemy knows they changed
        flag_modified(self, "processed_user_ids")

        if execution_state:
            self.execution_state = execution_state
            flag_modified(self, "execution_state")

        self.save(db)

    def mark_completed(self, db: Session) -> None:
        """Mark the execution as completed."""
        self.status = ExecutionLogStatus.complete
        self.completed_at = func.now()
        self.save(db)

    def mark_failed(self, db: Session, error_message: str) -> None:
        """Mark the execution as failed."""
        self.status = ExecutionLogStatus.error
        self.error_message = error_message
        self.completed_at = func.now()
        self.save(db)

    def can_resume(self) -> bool:
        """Check if this execution can be resumed."""
        return (
            self.status
            in [
                ExecutionLogStatus.in_processing,
                ExecutionLogStatus.awaiting_processing,
            ]
            and self.processed_user_ids is not None
        )

    def get_remaining_work(self) -> Dict[str, Any]:
        """Get information about remaining work for resumption."""
        # Ensure processed_user_ids is always a list
        raw_processed_user_ids = self.processed_user_ids or []
        processed_user_ids = (
            raw_processed_user_ids if isinstance(raw_processed_user_ids, list) else []
        )
        return {
            "processed_user_ids": processed_user_ids,
            "execution_state": self.execution_state or {},
            "processed_count": self.processed_recipients or 0,
            "successful_count": self.successful_communications,
            "failed_count": self.failed_communications,
        }



"""add_dismissed_to_monitor_task

Revision ID: e2cda8d6abc3
Revises: 0eef0016cf06
Create Date: 2025-09-17 23:37:49.678071

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e2cda8d6abc3"
down_revision = "0eef0016cf06"
branch_labels = None
depends_on = None


def upgrade():
    # Add the dismissed column to monitortask table
    op.add_column(
        "monitortask",
        sa.Column("dismissed", sa.Boolean(), nullable=False, server_default="f"),
    )


def downgrade():
    # Remove the dismissed column from monitortask table
    op.drop_column("monitortask", "dismissed")



"""
Database model for classification benchmark results.
"""

from __future__ import annotations

from datetime import datetime
from typing import Optional

from sqlalchemy import ARRAY, Column, ForeignKey, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import Query, RelationshipProperty, Session, relationship

from fides.api.db.base_class import Base, FidesBase
from fides.api.models.detection_discovery.core import MonitorConfig
from fides.api.models.sql_models import Dataset  # type: ignore[attr-defined]


class ClassificationBenchmark(Base, FidesBase):
    """
    Database model for storing classification benchmark results.

    This model stores the results of classification accuracy benchmarks that compare
    system-assigned classifications on staged resources with ground truth data from datasets.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "classification_benchmark"

    # Foreign key relationships
    monitor_config_key = Column(
        String,
        ForeignKey(MonitorConfig.key, ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    dataset_fides_key = Column(
        String,
        ForeignKey(Dataset.fides_key, ondelete="CASCADE"),
        nullable=False,
        index=True,
    )

    # Resource URNs (array of strings)
    resource_urns = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=list,
    )

    # Overall accuracy metrics stored as JSONB
    overall_metrics = Column(
        MutableDict.as_mutable(JSONB),
        nullable=True,
        server_default=None,
        default=None,
    )

    # Field-level accuracy details stored as JSONB array
    field_accuracy_details = Column(
        ARRAY(JSONB),
        nullable=False,
        server_default="{}",
        default=list,
    )

    # Status and messages for tracking benchmark execution
    status = Column(
        String,
        nullable=False,
        server_default="'pending'",
        default="pending",
        index=True,
    )
    messages = Column(
        ARRAY(String),
        nullable=False,
        server_default="{}",
        default=list,
    )

    # Relationships
    monitor_config: RelationshipProperty[MonitorConfig] = relationship(
        MonitorConfig,
        foreign_keys=[monitor_config_key],
        primaryjoin="ClassificationBenchmark.monitor_config_key == MonitorConfig.key",
    )

    dataset: RelationshipProperty[Dataset] = relationship(
        Dataset,
        foreign_keys=[dataset_fides_key],
        primaryjoin="ClassificationBenchmark.dataset_fides_key == Dataset.fides_key",
    )

    @classmethod
    def list_benchmarks(
        cls,
        db: Session,
        monitor_config_key: Optional[str] = None,
        dataset_fides_key: Optional[str] = None,
        created_after: Optional[datetime] = None,
        created_before: Optional[datetime] = None,
        status: Optional[str] = None,
    ) -> Query:
        """
        Get a filtered query for benchmarks.

        Args:
            db: Database session
            monitor_config_key: Filter by monitor config key
            dataset_fides_key: Filter by dataset fides key
            created_after: Filter by creation date (inclusive)
            created_before: Filter by creation date (exclusive)
            status: Filter by status
        Returns:
            Filtered query (not executed)
        """
        # Build the query with filters
        query = db.query(cls)

        # Apply filters
        if monitor_config_key:
            query = query.filter(cls.monitor_config_key == monitor_config_key)
        if dataset_fides_key:
            query = query.filter(cls.dataset_fides_key == dataset_fides_key)
        if created_after:
            query = query.filter(cls.created_at >= created_after)
        if created_before:
            query = query.filter(cls.created_at < created_before)
        if status:
            query = query.filter(cls.status == status)

        # Apply sorting
        query = query.order_by(cls.created_at.desc())

        return query



"""add_notice_display_order

Revision ID: 56fe6fad2d89
Revises: b2c3d4e5f6a7
Create Date: 2025-11-07 17:09:24.917188

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "56fe6fad2d89"
down_revision = "b2c3d4e5f6a7"
branch_labels = None
depends_on = None


def upgrade():
    """
    Add display_order column to experiencenotices and privacynotice tables.

    This allows:
    1. Experiences to control the order in which notices are displayed in the API response
    2. Parent notices to control the order in which child notices are displayed

    Both enable drag-and-drop reordering in the Admin UI.
    """
    # Add display_order to experiencenotices for ordering notices within an experience
    op.add_column(
        "experiencenotices",
        sa.Column("display_order", sa.Integer(), nullable=True),
    )

    # Add display_order to privacynotice for ordering child notices within a parent
    op.add_column(
        "privacynotice",
        sa.Column("display_order", sa.Integer(), nullable=True),
    )


def downgrade():
    """
    Remove display_order column from experiencenotices and privacynotice tables.
    """
    op.drop_column("experiencenotices", "display_order")
    op.drop_column("privacynotice", "display_order")



"""add_tagging_instructions_to_data_category

Revision ID: a55e12c2c2df
Revises: 65a1bc82ae09
Create Date: 2025-10-13 21:14:00.723464

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a55e12c2c2df"
down_revision = "65a1bc82ae09"
branch_labels = None
depends_on = None


def upgrade():
    """Add tagging_instructions column to ctl_data_categories table."""
    op.add_column(
        "ctl_data_categories",
        sa.Column("tagging_instructions", sa.Text(), nullable=True),
    )


def downgrade():
    """Remove tagging_instructions column from ctl_data_categories table."""
    op.drop_column("ctl_data_categories", "tagging_instructions")



"""Adding OAuth2 config for ConnectionConfig

Revision ID: 0eef0016cf06
Revises: 9e0dcbf67b9f
Create Date: 2025-09-12 01:06:26.838560

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "0eef0016cf06"
down_revision = "9e0dcbf67b9f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "oauth_config",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("connection_config_id", sa.String(), nullable=False),
        sa.Column(
            "grant_type",
            sa.Enum(
                "authorization_code",
                "client_credentials",
                "password",
                "implicit",
                name="oauthgranttype",
            ),
            nullable=False,
        ),
        sa.Column("token_url", sa.String(), nullable=False),
        sa.Column("scope", sa.String(), nullable=True),
        sa.Column("client_id", sa.String(), nullable=False),
        sa.Column(
            "client_secret", sqlalchemy_utils.types.StringEncryptedType, nullable=False
        ),
        sa.ForeignKeyConstraint(
            ["connection_config_id"],
            ["connectionconfig.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_oauth_config_id"), "oauth_config", ["id"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_oauth_config_id"), table_name="oauth_config")
    op.drop_table("oauth_config")
    # ### end Alembic commands ###



"""SQLAlchemy model for the v3 privacy_preferences table."""

from sqlalchemy import (  # type: ignore[attr-defined]
    BigInteger,
    Boolean,
    Column,
    DateTime,
    Identity,
    Text,
    text,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.cryptography.cryptographic_util import hash_value_with_salt
from fides.api.cryptography.identity_salt import get_identity_salt
from fides.api.db.base_class import Base
from fides.api.schemas.redis_cache import MultiValue
from fides.config import CONFIG


class PrivacyPreferences(Base):
    """
    Model for the v3 privacy_preferences table.
    This is used for the v3 privacy preferences API.

    This table stores privacy preference records with a partitioning scheme
    based on the is_latest flag, splitting records into current and historic partitions.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "privacy_preferences"

    # Override the default id column from Base (which is a String UUID)
    # with a BigInteger identity column for this table
    id = Column(
        BigInteger,  # type: ignore[arg-type]
        Identity(start=1, increment=1, always=True),
        primary_key=True,
    )

    # Searchable/queryable data stored as JSONB
    search_data = Column(JSONB, nullable=True)

    # Full record data stored as encrypted text (contains PII)
    record_data = Column(
        StringEncryptedType(
            type_in=Text(),
            key=CONFIG.security.app_encryption_key,
            engine=AesGcmEngine,
            padding="pkcs5",
        ),
        nullable=True,
    )

    # Partition key - determines if record goes to _current or _historic partition
    is_latest = Column(
        Boolean, nullable=False, server_default=text("false"), primary_key=True
    )

    # Override base class timestamp columns to match migration
    created_at = Column(
        DateTime(timezone=True), nullable=False, server_default=text("now()")
    )
    updated_at = Column(DateTime(timezone=True), nullable=True)

    @classmethod
    def hash_value(
        cls,
        value: MultiValue,
        encoding: str = "UTF-8",
    ) -> str:
        """Utility function to hash the value with a generated salt"""
        salt = get_identity_salt()
        value_str = str(value)
        hashed_value = hash_value_with_salt(
            value_str.encode(encoding),
            salt.encode(encoding),
        )
        return hashed_value



"""add digest execution model

Revision ID: b97e92b038d2
Revises: 3efe14d4469a
Create Date: 2025-10-01 16:42:41.900651

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "b97e92b038d2"
down_revision = "3efe14d4469a"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "digest_task_execution",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("action_type", sa.String(), nullable=False),
        sa.Column("digest_config_id", sa.String(), nullable=False),
        sa.Column("celery_task_id", sa.String(), nullable=True),
        sa.Column(
            "status",
            postgresql.ENUM(name="executionlogstatus", create_type=False),
            nullable=False,
        ),
        sa.Column("total_recipients", sa.Integer(), nullable=True),
        sa.Column("processed_recipients", sa.Integer(), nullable=False),
        sa.Column("successful_communications", sa.Integer(), nullable=False),
        sa.Column("failed_communications", sa.Integer(), nullable=False),
        sa.Column(
            "execution_state", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column(
            "processed_user_ids", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column("started_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("completed_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("last_checkpoint_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("error_message", sa.Text(), nullable=True),
        sa.ForeignKeyConstraint(
            ["digest_config_id"], ["digest_config.id"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_digest_task_execution_action_type"),
        "digest_task_execution",
        ["action_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_digest_task_execution_celery_task_id"),
        "digest_task_execution",
        ["celery_task_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_digest_task_execution_digest_config_id"),
        "digest_task_execution",
        ["digest_config_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_digest_task_execution_id"),
        "digest_task_execution",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_digest_task_execution_status"),
        "digest_task_execution",
        ["status"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_digest_task_execution_status"), table_name="digest_task_execution"
    )
    op.drop_index(
        op.f("ix_digest_task_execution_id"), table_name="digest_task_execution"
    )
    op.drop_index(
        op.f("ix_digest_task_execution_digest_config_id"),
        table_name="digest_task_execution",
    )
    op.drop_index(
        op.f("ix_digest_task_execution_celery_task_id"),
        table_name="digest_task_execution",
    )
    op.drop_index(
        op.f("ix_digest_task_execution_action_type"), table_name="digest_task_execution"
    )
    op.drop_table("digest_task_execution")
    # ### end Alembic commands ###



from typing import Any, Optional

from pydantic.v1.utils import deep_update

from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.policy import Policy as PolicySchema
from fides.api.task.conditional_dependencies.privacy_request.convenience_fields import (
    get_location_convenience_fields,
    get_policy_convenience_fields,
)
from fides.api.task.conditional_dependencies.privacy_request.schemas import (
    PrivacyRequestLocationConvenienceFields,
)
from fides.api.task.conditional_dependencies.util import (
    extract_nested_field_value,
    set_nested_value,
    transform_value_for_evaluation,
)


class PrivacyRequestDataTransformer:
    """
    Transforms a PrivacyRequest ORM object into a dictionary structure for condition evaluation.

    - This class extracts the fields referenced in conditions and builds a minimal nested structure
    for evaluation. Privacy request fields must be prefixed with the "privacy_request" namespace.
    - This class is used to get the privacy request data for condition evaluation.
        - The task.conditional_dependencies.evaluator expects the data to be in a nested dictionary structure
        and uses the operators defined in the task.conditional_dependencies.operators.py which have defined
        data type compatibility.

    Note: Privacy request relationships are eager loaded in the execute_request_tasks.py file.
    """

    def __init__(self, privacy_request: PrivacyRequest):
        self.privacy_request = privacy_request
        self.policy_data = self._transform_policy_object(privacy_request.policy)
        self.identity_data = self.privacy_request.get_persisted_identity()
        self.custom_privacy_request_fields_data = (
            self.privacy_request.get_persisted_custom_privacy_request_fields()
        )
        self.location_convenience_fields = get_location_convenience_fields(
            privacy_request.location
        )

    def to_evaluation_data(self, field_addresses: set[str]) -> dict[str, Any]:
        """
        Transforms the PrivacyRequest into an evaluation-ready dictionary structure.

        Args:
            field_addresses: Set of field addresses referenced in conditions
        """
        if not field_addresses:
            return {}

        result: dict[str, Any] = {"privacy_request": {}}
        field_paths: set[str] = {addr.replace(":", ".") for addr in field_addresses}

        # Extract each field dynamically based on its path
        for field_path in field_paths:
            value = self._extract_value_by_path(field_path)
            privacy_request_data: Optional[dict[str, Any]] = set_nested_value(
                field_path.split("."), value
            )
            if privacy_request_data:
                result = deep_update(result, privacy_request_data)

        return result

    def _extract_value_by_path(self, field_address: str) -> Any:
        """
        Extracts a value from PrivacyRequest by following the field address path.

        Args:
            field_address: The field address path (e.g., "privacy_request.location", "privacy_request.policy.rules")

        """
        # first part will always be privacy_request
        parts: list[str] = field_address.split(".")[1:]
        current: Any = self.privacy_request
        is_labeled_field = False

        # Handle location convenience fields - these are derived fields, not direct attributes
        # Check if we're accessing a location convenience field before trying to extract from privacy_request

        if len(parts) == 1 and parts[0] in [
            e.value for e in PrivacyRequestLocationConvenienceFields
        ]:
            return self.location_convenience_fields.get(parts[0])

        # Route to the appropriate data source based on the first path component
        if parts[0] == "policy":
            current = self.policy_data
            parts.pop(0)
        elif parts[0] == "identity":
            current = self.identity_data
            # Identity fields may contain LabeledIdentity objects with label/value structure
            is_labeled_field = True
            parts.pop(0)
        elif parts[0] == "custom_privacy_request_fields":
            current = self.custom_privacy_request_fields_data
            # Custom fields are stored as {"label": ..., "value": ...} dicts
            is_labeled_field = True
            parts.pop(0)

        current = extract_nested_field_value(current, parts)

        # If we started with an Identity object and didn't extract any field from it
        # (current is still the identity object), return None
        # Otherwise, transform the value (which could be a BaseModel like LabeledIdentity)
        if current is self.identity_data:
            return None

        # Transform the value first (converts LabeledIdentity to dict, handles other types)
        current = transform_value_for_evaluation(current)

        # For labeled fields (custom privacy request fields and identity fields with labels),
        # extract just the value from the {"label": ..., "value": ...} structure
        if is_labeled_field and isinstance(current, dict) and "value" in current:
            current = current["value"]

        return current

    def _transform_policy_object(self, policy: Policy) -> dict[str, Any]:
        """
        Transforms a Policy ORM object to a dictionary using existing Pydantic schema
        and adds convenience fields.

        Args:
            policy: The Policy ORM object
        """
        if not policy:
            return {}
        extra_fields = get_policy_convenience_fields(policy)

        # Convert Policy ORM object to dictionary using Pydantic schema
        policy_dict = PolicySchema.model_validate(policy).model_dump()
        policy_dict.update(extra_fields)
        return policy_dict



"""migrate field type from enum to string to support custom taxonomies

Revision ID: 795f46f656c0
Revises: 303287c70600
Create Date: 2025-11-18 18:30:00.000000

"""

import sqlalchemy as sa
from alembic import op
from loguru import logger

# revision identifiers, used by Alembic.
revision = "795f46f656c0"
down_revision = "303287c70600"
branch_labels = None
depends_on = None

# Original enum values used by AllowedTypes
ALLOWED_TYPES = (
    "string",
    "string[]",
)

# Name for the PostgreSQL enum type that backs AllowedTypes
PG_ENUM_NAME = "allowedtypes"


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    # Alter field_type column from Enum -> String
    # Use a PostgreSQL USING clause to safely cast enum -> text
    op.alter_column(
        "plus_custom_field_definition",
        "field_type",
        existing_type=sa.Enum(*ALLOWED_TYPES, name=PG_ENUM_NAME),
        type_=sa.String(),
        postgresql_using="field_type::text",
        nullable=False,
    )

    # Data migration: Convert old enum value 'string_list' to 'string[]'
    op.execute(
        """
        UPDATE plus_custom_field_definition
        SET field_type = 'string[]'
        WHERE field_type = 'string_list'
        """
    )

    # Drop the unused enum type after the column has been converted
    op.execute(f'DROP TYPE IF EXISTS "{PG_ENUM_NAME}"')
    # ### end Alembic commands ###


def downgrade():
    """
    This migration does not support downgrades.
    Converting string back to enum would fail if taxonomy keys are present.
    """
    logger.info("Conversion of str to enum AllowedTypes is unsupported.")



"""
Pure HTTP utility for executing polling requests.

This module contains low-level HTTP request execution for async DSR polling,
with no business logic or orchestration dependencies.
"""

from typing import Any, Dict, Optional

from requests import Response

from fides.api.common_exceptions import PrivacyRequestError
from fides.api.schemas.saas.async_polling_configuration import (
    PollingResultRequest,
    PollingStatusRequest,
)
from fides.api.service.connectors.saas.authenticated_client import AuthenticatedClient
from fides.api.util.saas_util import map_param_values


class PollingRequestHandler:
    """
    Pure HTTP utility for executing polling requests.

    Handles status checks and result retrieval with no business logic dependencies.
    """

    def __init__(
        self,
        status_request: PollingStatusRequest,
        result_request: Optional[PollingResultRequest] = None,
    ):
        self.status_request = status_request
        self.result_request = result_request

    def get_status_response(
        self,
        client: AuthenticatedClient,
        param_values: Dict[str, Any],
    ) -> Response:
        """Execute HTTP status request and return raw response."""
        if not self.status_request:
            raise PrivacyRequestError(
                "status_request is not configured in the async polling configuration"
            )

        prepared_status_request = map_param_values(
            action="status",
            context="polling request",
            current_request=self.status_request,
            param_values=param_values,
        )

        response: Response = client.send(prepared_status_request)

        if not response.ok:
            raise PrivacyRequestError(
                f"Status request failed with status code {response.status_code}: {response.text}"
            )

        return response

    def get_result_response(
        self,
        client: AuthenticatedClient,
        param_values: Dict[str, Any],
    ) -> Response:
        """Execute HTTP result request and return raw response."""
        if not self.result_request:
            raise PrivacyRequestError(
                "result_request is not configured in the async polling configuration"
            )

        prepared_result_request = map_param_values(
            action="result",
            context="polling request",
            current_request=self.result_request,
            param_values=param_values,
        )

        response: Response = client.send(prepared_result_request)

        if not response.ok:
            raise PrivacyRequestError(
                f"Result request failed with status code {response.status_code}: {response.text}"
            )

        return response



from datetime import datetime
from enum import Enum
from typing import Any, Dict, Optional, Tuple

from fides.api.common_exceptions import NoSuchConnectionTypeSecretSchemaError
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.event_audit import EventAuditType
from fides.api.schemas.connection_configuration.connection_oauth_config import (
    OAuthConfigSchema,
)
from fides.api.util.connection_type import get_connection_type_secret_schema
from fides.api.util.masking_util import mask_sensitive_fields


def normalize_value(value: Any) -> Any:
    """
    Normalize values for consistent comparison and serialization.

    - Converts datetime objects to ISO strings
    - Converts enums to their string values
    - Converts lists of enums to lists of string values
    - Treats None and empty strings as equivalent (both become None)
    """
    # Treat None and empty strings as equivalent
    if value == "" or value is None:
        return None
    if isinstance(value, datetime):
        return value.isoformat()
    if isinstance(value, Enum):
        return value.value
    if (
        isinstance(value, list) and value and hasattr(value[0], "value")
    ):  # List of enums
        return [item.value for item in value]
    return value


def _get_saas_connector_type(connection_config: ConnectionConfig) -> Optional[str]:
    """
    Helper function to extract SaaS connector type from connection config.

    Args:
        connection_config: The connection configuration

    Returns:
        SaaS connector type if available, None otherwise
    """
    connection_type = connection_config.connection_type.value  # type: ignore[attr-defined]
    if connection_type == "saas" and connection_config.saas_config:
        try:
            saas_config = connection_config.get_saas_config()
            if saas_config:
                return saas_config.type
        except Exception:
            # If SaaS config is invalid, try to get type directly from raw config
            if (
                isinstance(connection_config.saas_config, dict)
                and "type" in connection_config.saas_config
            ):
                return connection_config.saas_config["type"]
    return None


def _create_connection_event_details(
    connection_config: ConnectionConfig,
    operation_type: str,
    changed_fields: Optional[set] = None,
) -> Dict[str, Any]:
    """
    Create standardized event details for connection audit events.

    Args:
        connection_config: The connection configuration
        operation_type: Type of operation (create, update, delete)
        changed_fields: Set of field names that changed (None means include all fields)

    Returns:
        Standardized event details dictionary
    """
    event_details: dict[str, Any] = {
        "operation_type": operation_type,
    }

    # Define fields to exclude from audit details
    excluded_fields = {
        "id",
        "created_at",
        "updated_at",
        "secrets",
    }

    # Dynamically add connection config fields (only changed fields if specified)
    configuration_changes: dict[str, Any] = {}
    for column in connection_config.__table__.columns:
        field_name = column.name
        if field_name not in excluded_fields:
            # If changed_fields is specified, only include fields that changed
            if changed_fields is not None and field_name not in changed_fields:
                continue

            value = getattr(connection_config, field_name)

            configuration_changes[field_name] = normalize_value(value)

    event_details["configuration_changes"] = configuration_changes
    return event_details


def generate_connection_audit_event_details(
    event_type: EventAuditType,
    connection_config: ConnectionConfig,
    description: Optional[str] = None,
    changed_fields: Optional[set] = None,
) -> Tuple[Dict[str, Any], str]:
    """
    Create an audit event for connection operations.

    Args:
        event_type: Type of audit event
        connection_config: The connection configuration
        description: Human-readable description
        changed_fields: Set of field names that changed (None means include all fields)

    Returns:
        Created EventAudit instance
    """
    # Determine operation type from event type
    operation_type = event_type.value.split(".")[
        -1
    ]  # e.g., "created", "updated", "deleted"

    # Create standardized event details
    if operation_type == "deleted":
        event_details = {
            "operation_type": operation_type,
        }
    else:
        event_details = _create_connection_event_details(
            connection_config,
            operation_type,
            changed_fields,
        )

    # Generate description if not provided
    connection_type = connection_config.connection_type.value  # type: ignore[attr-defined]
    connector_type = _get_saas_connector_type(connection_config)

    description = f"Connection {operation_type}: {connector_type if connector_type else connection_type} connection '{connection_config.key}'"

    return event_details, description


def generate_connection_secrets_event_details(
    event_type: EventAuditType,
    connection_config: ConnectionConfig,
    secrets_modified: Dict[str, Any],
) -> Tuple[Dict[str, Any], str]:
    """
    Create an audit event for connection secrets operations.

    Args:
        event_type: Type of audit event (should be a connection.secrets.* event)
        connection_config: The connection configuration
        secrets_modified: Dict of secret field names and values that were modified
        changed_fields: Set of field names that changed (None means include all fields)
        description: Human-readable description

    Returns:
        Created EventAudit instance
    """
    # Determine operation type from event type
    operation_type = event_type.value.split(".")[
        -1
    ]  # e.g., "created", "updated", "deleted"

    # Mask the secrets using the existing masking utility
    saas_config = connection_config.get_saas_config()
    connection_type = connection_config.connection_type.value  # type: ignore[attr-defined]
    connection_type = (
        saas_config.type
        if connection_type == "saas" and saas_config
        else connection_type
    )

    # Get the secret schema for this connection type
    try:
        secret_schema = None

        # For HTTPS connections with OAuth config, check if we should use OAuth schema
        if connection_type == "https" and connection_config.oauth_config:
            # Get OAuth fields dynamically from the schema
            oauth_schema = OAuthConfigSchema.model_json_schema()
            oauth_fields = set(oauth_schema.get("properties", {}).keys())
            secrets_keys = set(secrets_modified.keys())

            # If we're updating OAuth fields, use OAuth schema
            if secrets_keys.intersection(oauth_fields):
                secret_schema = oauth_schema

        # If we haven't set a schema yet, use the connection type's secret schema
        if secret_schema is None:
            secret_schema = get_connection_type_secret_schema(
                connection_type=connection_type
            )

        # Use existing masking function
        masked_secrets = mask_sensitive_fields(secrets_modified, secret_schema)
    except NoSuchConnectionTypeSecretSchemaError:
        # if there is no schema, mask every secret
        masked_secrets = {key: "**********" for key in secrets_modified}

    # Create event details with masked secret values
    event_details: dict[str, Any] = {
        "secrets": masked_secrets,
    }

    # Add SaaS connector type if applicable
    connector_type = _get_saas_connector_type(connection_config)
    if connector_type:
        event_details["saas_connector_type"] = connector_type

    # Generate description
    secret_names = ", ".join(secrets_modified) if secrets_modified else "secrets"

    if connector_type:
        description = f"Connection secrets {operation_type}: {connector_type} connection '{connection_config.key}' - {secret_names}"
    else:
        description = f"Connection secrets {operation_type}: {connection_type} connection '{connection_config.key}' - {secret_names}"

    return event_details, description



"""adds new experience configs for vendor asset disclosure

Revision ID: 3efe14d4469a
Revises: f108fa05c579
Create Date: 2025-09-09 11:19:24.060587

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "3efe14d4469a"
down_revision = "f108fa05c579"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "experienceconfigtemplate",
        sa.Column(
            "asset_disclosure_include_types",
            postgresql.ARRAY(sa.String()),
            nullable=True,
        ),
    )
    op.add_column(
        "experienceconfigtemplate",
        sa.Column(
            "allow_vendor_asset_disclosure",
            sa.Boolean(),
            server_default="f",
            nullable=False,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "asset_disclosure_include_types",
            postgresql.ARRAY(sa.String()),
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "allow_vendor_asset_disclosure",
            sa.Boolean(),
            server_default="f",
            nullable=False,
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("allow_vendor_asset_disclosure", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "asset_disclosure_include_types",
            postgresql.ARRAY(sa.String()),
            nullable=True,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyexperienceconfighistory", "asset_disclosure_include_types")
    op.drop_column("privacyexperienceconfighistory", "allow_vendor_asset_disclosure")
    op.drop_column("privacyexperienceconfig", "allow_vendor_asset_disclosure")
    op.drop_column("privacyexperienceconfig", "asset_disclosure_include_types")
    op.drop_column("experienceconfigtemplate", "allow_vendor_asset_disclosure")
    op.drop_column("experienceconfigtemplate", "asset_disclosure_include_types")
    # ### end Alembic commands ###



from typing import Iterable

from loguru import logger
from packaging.version import Version
from packaging.version import parse as parse_version
from sqlalchemy.orm import Session

from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.saas_template_dataset import SaasTemplateDataset
from fides.api.schemas.saas.connector_template import ConnectorTemplate
from fides.api.schemas.saas.saas_config import SaaSConfig
from fides.api.service.connectors.saas.connector_registry_service import (
    ConnectorRegistry,
)
from fides.api.util.saas_util import load_config_from_string, load_dataset_from_string
from fides.service.connection.connection_service import ConnectionService
from fides.service.event_audit_service import EventAuditService


def update_saas_configs(db: Session) -> None:
    """
    Updates SaaS config instances currently in the DB if to the
    corresponding template in the registry are found.

    Effectively an "update script" for SaaS config instances,
    to be run on server bootstrap.
    """
    event_audit_service = EventAuditService(db)
    saas_connection_service = ConnectionService(db, event_audit_service)
    for connector_type in ConnectorRegistry.connector_types():
        logger.debug(
            "Determining if any updates are needed for connectors of type {} based on templates...",
            connector_type,
        )
        template: ConnectorTemplate = ConnectorRegistry.get_connector_template(  # type: ignore
            connector_type
        )

        # Store the original template dataset (with placeholders) instead of the modified version
        template_dataset_json = load_dataset_from_string(template.dataset)

        SaasTemplateDataset.get_or_create(
            db=db,
            connector_type=connector_type,
            dataset_json=template_dataset_json,
        )

        saas_config = SaaSConfig(**load_config_from_string(template.config))
        template_version: Version = parse_version(saas_config.version)
        connection_configs: Iterable[ConnectionConfig] = ConnectionConfig.filter(
            db=db,
            conditions=(ConnectionConfig.saas_config["type"].astext == connector_type),
        ).all()
        for connection_config in connection_configs:
            saas_config_instance = SaaSConfig.model_validate(
                connection_config.saas_config
            )
            if parse_version(saas_config_instance.version) < template_version:
                logger.info(
                    "Updating SaaS config instance '{}' of type '{}' as its version, {}, was found to be lower than the template version {}",
                    saas_config_instance.fides_key,
                    connector_type,
                    saas_config_instance.version,
                    template_version,
                )
                try:
                    saas_connection_service.update_saas_instance(
                        connection_config,
                        template,
                        saas_config_instance,
                    )
                except Exception:
                    logger.exception(
                        "Encountered error attempting to update SaaS config instance {}",
                        saas_config_instance.fides_key,
                    )



"""adds experience config for delete cookie based on host domain

Revision ID: 65a1bc82ae09
Revises: 4bfbeff34611
Create Date: 2025-10-07 11:57:39.614312

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "65a1bc82ae09"
down_revision = "4bfbeff34611"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "experienceconfigtemplate",
        sa.Column(
            "cookie_deletion_based_on_host_domain",
            sa.Boolean(),
            server_default="t",
            nullable=False,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "cookie_deletion_based_on_host_domain",
            sa.Boolean(),
            server_default="t",
            nullable=False,
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("cookie_deletion_based_on_host_domain", sa.Boolean(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column(
        "privacyexperienceconfighistory", "cookie_deletion_based_on_host_domain"
    )
    op.drop_column("privacyexperienceconfig", "cookie_deletion_based_on_host_domain")
    op.drop_column("experienceconfigtemplate", "cookie_deletion_based_on_host_domain")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import Any, Dict, List, Optional, Type

from pydantic import ConfigDict
from sqlalchemy import Boolean, Column, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import RelationshipProperty, Session, relationship

from fides.api.db.base_class import Base
from fides.api.models.property import Property
from fides.api.schemas.messaging.messaging import MessagingActionType

# Provides default values for initializing the database or replacing deleted values for messaging templates.
# Note: There are additional MessagingActionTypes that are internally used but are not exposed for user customization.
DEFAULT_MESSAGING_TEMPLATES: Dict[str, Any] = {
    MessagingActionType.SUBJECT_IDENTITY_VERIFICATION.value: {
        "label": "Subject identity verification",
        "content": {
            "subject": "Your one-time code is __CODE__",
            "body": "Your privacy request verification code is __CODE__. Please return to the Privacy Center and enter the code to continue. This code will expire in __MINUTES__ minutes.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_RECEIPT.value: {
        "label": "Privacy request received",
        "content": {
            "subject": "Your privacy request has been received",
            "body": "Your privacy request has been received. We will get back to you shortly.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_REVIEW_APPROVE.value: {
        "label": "Privacy request approved",
        "content": {
            "subject": "Your privacy request has been approved",
            "body": "Your privacy request has been approved and is currently processing.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_REVIEW_DENY.value: {
        "label": "Privacy request denied",
        "content": {
            "subject": "Your privacy request has been denied",
            "body": "Your privacy request has been denied. __DENIAL_REASON__.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_COMPLETE_ACCESS.value: {
        "label": "Access request completed",
        "content": {
            "subject": "Your data is ready to be downloaded",
            "body": "Your access request has been completed and can be downloaded at __DOWNLOAD_LINK__. For security purposes, this secret link will expire in __DAYS__ days.",
        },
    },
    MessagingActionType.PRIVACY_REQUEST_COMPLETE_DELETION.value: {
        "label": "Erasure request completed",
        "content": {
            "subject": "Your data has been deleted",
            "body": "Your erasure request has been completed.",
        },
    },
    MessagingActionType.MANUAL_TASK_DIGEST.value: {
        "label": "Manual task digest",
        "content": {
            "subject": "Weekly DSR Summary from __ORGANIZATION_NAME__",
            "body": "This is your weekly summary of open data subject requests tasks from __ORGANIZATION_NAME__ that require your attention. Please review the tasks in our Privacy Center to ensure timely completion.",
        },
    },
    MessagingActionType.EXTERNAL_USER_WELCOME.value: {
        "label": "External user welcome",
        "content": {
            "subject": "Welcome to our Privacy Center",
            "body": "An account has been created for you by the privacy team at __ORG_NAME__ so that you can assist them in responding to data subject requests, as required by the GDPR and many US state laws, for users that have data in your systems. A portal is available to view and take action on these requests at: __PORTAL_LINK__. Through this portal you can upload or provide user data for users who have requested a copy of their information and attest that users have been removed from your system in the case of an erasure request. The provided link is valid for 30 days and you will be asked to verify your email address each time you access the portal. If you have any questions, please reach out first to your contacts at: __ORG_NAME__",
        },
    },
}


def _link_properties_to_messaging_template(
    db: Session,
    properties: List[Dict[str, Any]],
    messaging_template: MessagingTemplate,
) -> Optional[List[Property]]:
    """
    Link supplied properties to MessagingTemplate and unlink any properties not supplied.
    """
    new_properties = []
    if len(properties) > 0:
        new_properties = (
            db.query(Property)
            .filter(Property.id.in_([prop["id"] for prop in properties]))
            .all()
        )
    messaging_template.properties = new_properties
    messaging_template.save(db)
    return messaging_template.properties


class MessagingTemplate(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "messaging_template"

    type = Column(String, index=True, nullable=False)
    content = Column(MutableDict.as_mutable(JSONB), nullable=False)
    properties: RelationshipProperty[List[Property]] = relationship(
        "Property",
        secondary="messaging_template_to_property",
        back_populates="messaging_templates",
        lazy="selectin",
    )
    is_enabled = Column(Boolean, default=False, nullable=False)
    model_config = ConfigDict(from_attributes=True)

    @classmethod
    def create(
        cls: Type[MessagingTemplate],
        db: Session,
        *,
        data: dict[str, Any],
        check_name: bool = True,
    ) -> MessagingTemplate:
        """
        Creates a Messaging Template, allows linking properties
        """
        properties = data.pop("properties", [])
        messaging_template: MessagingTemplate = super().create(
            db=db, data=data, check_name=check_name
        )
        # Link Properties to this Messaging Template via the MessagingTemplateToProperty table
        _link_properties_to_messaging_template(db, properties, messaging_template)

        return messaging_template

    def update(self, db: Session, *, data: dict[str, Any]) -> MessagingTemplate:
        """
        Updates a Messaging Template, allows linking properties
        """
        properties = data.pop("properties", [])
        super().update(db=db, data=data)

        # Link Properties to this Messaging Template via the MessagingTemplateToProperty table
        _link_properties_to_messaging_template(db, properties, self)
        return self



from enum import Enum
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field

from fides.api.schemas.base_class import NoValidationSchema


class OAuthGrantType(str, Enum):
    """OAuth2 grant types supported by the system"""

    client_credentials = "client_credentials"


class OAuthConfigSchema(BaseModel):
    """Schema for OAuth2 configuration used in API requests"""

    grant_type: Optional[OAuthGrantType] = None
    token_url: Optional[str] = None
    scope: Optional[str] = None
    client_id: Optional[str] = None
    client_secret: Optional[str] = Field(json_schema_extra={"sensitive": True})

    model_config = ConfigDict(
        from_attributes=True, use_enum_values=True, extra="forbid"
    )


class OAuthConfigResponse(BaseModel):
    """Schema for OAuth2 configuration used in API responses with masked sensitive fields"""

    grant_type: OAuthGrantType
    token_url: str
    scope: Optional[str] = None
    client_id: str
    client_secret: str = "**********"  # Always masked in responses

    model_config = ConfigDict(from_attributes=True, use_enum_values=True)


class OAuthConfigDocsSchema(OAuthConfigSchema, NoValidationSchema):
    """OAuth2 Config Schema for API Docs"""



from typing import Dict

from sqlalchemy import Column, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import Session

from fides.api.db.base_class import Base


class SaasTemplateDataset(Base):
    """
    Stores the latest template dataset for each SaaS connection type.

    This table maintains a record of the template datasets that come with
    Fides for each SaaS connector type. It's used as a baseline for
    detecting customer modifications to datasets.
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "saas_template_dataset"

    connection_type = Column(String, nullable=False, unique=True, index=True)
    dataset_json = Column(JSONB, nullable=False)

    def __repr__(self) -> str:
        return f"<SaasTemplateDataset(connection_type='{self.connection_type}')>"

    @classmethod
    def get_or_create(  # type: ignore[override]
        cls,
        db: Session,
        connector_type: str,
        dataset_json: Dict,
    ) -> tuple[bool, "SaasTemplateDataset"]:
        """
        Get existing SaasTemplateDataset by connector_type or create if it doesn't exist.

        Args:
            db: Database session
            connector_type: The connection type identifier
            dataset_json: The dataset JSON to use if creating

        Returns:
            Existing or newly created SaasTemplateDataset instance
        """
        created = False
        template_dataset = cls.get_by(
            db=db, field="connection_type", value=connector_type
        )

        if not template_dataset:
            template_dataset = cls.create(
                db=db,
                data={
                    "connection_type": connector_type,
                    "dataset_json": dataset_json,
                },
            )
            created = True

        return created, template_dataset



from enum import Enum

from sqlalchemy import Boolean, Column, String, Text
from sqlalchemy.ext.declarative import declared_attr

from fides.api.db.base_class import Base, FidesBase
from fides.api.db.util import EnumColumn


class IdentityDefinitionType(str, Enum):
    """Enum for the type of identity"""

    EMAIL = "email"
    PHONE_NUMBER = "phone_number"
    UUID = "uuid"
    STRING = "string"
    INTEGER = "integer"


class IdentityDefinition(Base):
    """
    Model for identity definitions in Fides. This isn't for specific identity values,
    but for the types of identities that can be used in Fides.

    For example:
    ```json
    {
        "identity_key": "customer_id",
        "name": "Customer ID",
        "description": "The unique identifier for the customer",
        "type": "string"
    }
    ```
    """

    @declared_attr
    def __tablename__(self) -> str:
        return "identity_definition"

    # Overriding the id definition from Base so we don't treat this as the primary key
    id = Column(
        String(255),
        nullable=False,
        index=False,
        unique=True,
        default=FidesBase.generate_uuid,
    )

    # Primary key
    identity_key = Column(String(255), primary_key=True)

    # Schema definition
    name = Column(String(255), nullable=False)
    description = Column(Text, nullable=True)
    type = Column(
        EnumColumn(
            IdentityDefinitionType,
            native_enum=False,
            values_callable=lambda x: [
                i.value for i in x
            ],  # allows enum _values_ to be stored rather than name
        ),
        nullable=False,
    )
    created_by = Column(String(255), nullable=True)
    is_default = Column(Boolean, nullable=False, server_default="false")



from abc import abstractmethod
from typing import Any, Dict, List

from sqlalchemy.orm import Session

from fides.api.models.privacy_request.request_task import AsyncTaskType, RequestTask
from fides.api.service.connectors.query_configs.saas_query_config import SaaSQueryConfig
from fides.api.service.connectors.saas.authenticated_client import AuthenticatedClient
from fides.api.service.strategy import Strategy
from fides.api.util.collection_util import Row


class AsyncDSRStrategy(Strategy):
    """Abstract base class for async DSR strategies"""

    type: AsyncTaskType
    session: Session

    @abstractmethod
    def async_retrieve_data(
        self,
        client: AuthenticatedClient,
        request_task_id: str,
        query_config: SaaSQueryConfig,
        input_data: Dict[str, List[Any]],
    ) -> List[Row]:
        """
        Execute async retrieve data.
        """

    @abstractmethod
    def async_mask_data(
        self,
        client: AuthenticatedClient,
        request_task_id: str,
        query_config: SaaSQueryConfig,
        rows: List[Row],
    ) -> int:
        """
        Execute async mask data.
        """

    def _get_request_task(self, request_task_id: str) -> RequestTask:
        """Get request task by ID or raise ValueError if not found."""
        request_task = (
            self.session.query(RequestTask)
            .filter(RequestTask.id == request_task_id)
            .first()
        )
        if not request_task:
            raise ValueError(f"RequestTask with ID {request_task_id} not found")
        return request_task



from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field, model_validator

from fides.api.schemas.saas.saas_config import SaaSRequest
from fides.api.schemas.saas.strategy_configuration import StrategyConfiguration
from fides.api.util.collection_util import Row


class SupportedDataType(Enum):
    """Supported data types for polling async DSR result requests."""

    # Structured data types that can be parsed into rows
    json = "json"  # Parsed into List[Row] from JSON response
    csv = "csv"  # Parsed into List[Row] from CSV response
    # Binary/non-parseable data stored as raw bytes
    attachment = "attachment"  # Binary files (.zip, .pdf, .xml, etc.) stored as bytes


class PollingResultType(Enum):
    """Types of results from async polling operations."""

    rows = "rows"  # Structured data parsed into List[Row]
    attachment = "attachment"  # Binary file data stored as bytes


class PollingResult(BaseModel):
    """
    Flexible result container for async polling operations.
    Handles both structured data and file attachments.
    """

    data: Union[List[Row], bytes]
    result_type: PollingResultType
    metadata: Dict[str, Any] = Field(default_factory=dict)


class PollingStatusRequest(SaaSRequest):
    """
    Extended SaaSRequest for checking async job status.
    Uses request_override for custom status checking logic or standard fields for simple cases.
    """

    status_path: Optional[str] = None
    status_completed_value: Optional[Union[str, bool, int]] = None

    @model_validator(mode="after")
    def validate_status_fields(self) -> "PollingStatusRequest":
        """Ensure required fields are present unless using an override."""
        if self.request_override:
            return self

        if not self.status_path:
            raise ValueError("status_path is required when request_override is not set")
        if self.status_completed_value is None:
            raise ValueError(
                "status_completed_value is required when request_override is not set"
            )
        return self


class PollingResultRequest(SaaSRequest):
    """
    Extended SaaSRequest for retrieving async job results.
    Uses request_override for custom result retrieval or standard HTTP request for simple cases.
    Data type is automatically inferred from response.
    """

    result_path: Optional[str] = None


class AsyncPollingConfiguration(StrategyConfiguration):
    """
    Simplified configuration for polling async DSR requests.
    The main read request serves as the initial request.
    """

    status_request: PollingStatusRequest
    # result_request is optional for delete/update operations
    result_request: Optional[PollingResultRequest] = None



from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import ValidationError
from sqlalchemy.orm import Session

from fides.api.common_exceptions import NoSuchStrategyException
from fides.api.common_exceptions import ValidationError as FidesopsValidationError
from fides.api.schemas.saas.async_polling_configuration import AsyncPollingConfiguration
from fides.api.service.async_dsr.strategies.async_dsr_strategy import AsyncDSRStrategy
from fides.api.service.async_dsr.strategies.async_dsr_strategy_callback import (
    AsyncCallbackStrategy,
)
from fides.api.service.async_dsr.strategies.async_dsr_strategy_polling import (
    AsyncPollingStrategy,
)


class SupportedAsyncDSRStrategies(Enum):
    """
    The supported async DSR strategies for handling asynchronous data subject requests.
    """

    callback = AsyncCallbackStrategy
    polling = AsyncPollingStrategy

    @classmethod
    def __contains__(cls, item: str) -> bool:
        try:
            cls[item]
        except KeyError:
            return False

        return True


def get_strategy(
    strategy_name: str,
    session: Session,
    configuration: Optional[Dict[str, Any]] = None,
) -> AsyncDSRStrategy:
    """
    Returns the enhanced strategy given the name, session, and configuration.
    Raises NoSuchStrategyException if the strategy does not exist
    """
    if not SupportedAsyncDSRStrategies.__contains__(strategy_name):
        valid_strategies = ", ".join(get_strategy_names())
        raise NoSuchStrategyException(
            f"Strategy '{strategy_name}' does not exist. Valid strategies are [{valid_strategies}]"
        )

    strategy_class = SupportedAsyncDSRStrategies[strategy_name].value

    if strategy_name == "polling":
        # Polling strategy requires configuration
        if not configuration:
            raise FidesopsValidationError("Configuration required for polling strategy")
        try:
            strategy_config = AsyncPollingConfiguration(**configuration)
            return strategy_class(session, strategy_config)
        except ValidationError as e:
            raise FidesopsValidationError(message=str(e))
    elif strategy_name == "callback":
        # Callback strategy only requires session
        return strategy_class(session)

    raise NoSuchStrategyException(f"Unsupported strategy: {strategy_name}")


def get_strategy_names() -> List[str]:
    """Returns all supported async DSR strategies"""
    return [s.name for s in SupportedAsyncDSRStrategies]



"""Create new Sub Request Table

Revision ID: 7db29f9cd77b
Revises: b97e92b038d2
Create Date: 2025-09-16 14:00:16.282996

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy_utils.types.encrypted.encrypted_type import (
    AesGcmEngine,
    StringEncryptedType,
)

from fides.api.db.base_class import JSONTypeOverride
from fides.config import CONFIG

# revision identifiers, used by Alembic.
revision = "7db29f9cd77b"
down_revision = "b97e92b038d2"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "request_task_sub_request",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("request_task_id", sa.String(length=255), nullable=False),
        sa.Column(
            "param_values",
            StringEncryptedType(
                type_in=JSONTypeOverride,
                key=CONFIG.security.app_encryption_key,
                engine=AesGcmEngine,
                padding="pkcs5",
            ),
            nullable=False,
        ),
        sa.Column("status", sa.String(), nullable=False),
        sa.Column(
            "access_data",
            StringEncryptedType(
                type_in=JSONTypeOverride,
                key=CONFIG.security.app_encryption_key,
                engine=AesGcmEngine,
                padding="pkcs5",
            ),
            nullable=True,
        ),
        sa.Column("rows_masked", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["request_task_id"],
            ["requesttask.id"],
            name="request_task_sub_request_request_task_id_fkey",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_request_task_sub_request_id"),
        "request_task_sub_request",
        ["id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_request_task_sub_request_request_task_id"),
        "request_task_sub_request",
        ["request_task_id"],
        unique=False,
    )


def downgrade():
    op.drop_index(
        op.f("ix_request_task_sub_request_request_task_id"),
        table_name="request_task_sub_request",
    )
    op.drop_index(
        op.f("ix_request_task_sub_request_id"), table_name="request_task_sub_request"
    )
    op.drop_table("request_task_sub_request")



"""
Contains all of the logic related to the database including connections, setup, teardown, etc.
"""

import os
import subprocess
from os import path
from pathlib import Path
from typing import Literal, Optional, Tuple

from alembic import command, script
from alembic.config import Config
from alembic.runtime import migration
from loguru import logger as log
from sqlalchemy import text
from sqlalchemy.orm import Session
from sqlalchemy.schema import SchemaItem
from sqlalchemy_utils.functions import create_database, database_exists
from sqlalchemy_utils.types.encrypted.encrypted_type import InvalidCiphertextError

from fides.api.db.base import Base  # type: ignore[attr-defined]
from fides.api.db.seed import load_default_resources
from fides.api.db.session import get_db_engine
from fides.api.util.errors import get_full_exception_name

DatabaseHealth = Literal["healthy", "unhealthy", "needs migration"]

# Tables to exclude from migration auto-generation (e.g., tables without SQLAlchemy models)
EXCLUDED_TABLES = {
    "privacy_preferences_current",
    "privacy_preferences_historic",
}


def include_object(
    object: SchemaItem,  # pylint: disable=redefined-builtin
    name: str,
    type_: str,
    reflected: bool,
    compare_to: SchemaItem,
) -> bool:
    """Filter out excluded tables from migration comparison and auto-generation."""
    if type_ == "table" and name in EXCLUDED_TABLES:
        return False
    return True


def get_alembic_config(database_url: str) -> Config:
    """
    Do lots of magic to make alembic work programmatically from the CLI.
    """

    migrations_dir = path.dirname(path.abspath(__file__))
    directory = path.join(migrations_dir, "../alembic/migrations")
    config = Config(path.join(migrations_dir, "../alembic/alembic.ini"))
    config.set_main_option("script_location", directory.replace("%", "%%"))
    # Avoids invalid interpolation syntax errors if % in string
    config.set_main_option("sqlalchemy.url", database_url.replace("%", "%%"))
    return config


def upgrade_db(alembic_config: Config, revision: str = "head") -> None:
    """Upgrade the database to the specified migration revision."""
    log.info(f"Running database upgrade to revision {revision}")
    command.upgrade(alembic_config, revision)


def init_db(database_url: str) -> None:
    """
    Runs the migrations and creates all of the database objects.
    """
    alembic_config = get_alembic_config(database_url)
    upgrade_db(alembic_config)


def downgrade_db(alembic_config: Config, revision: str = "head") -> None:
    """Downgrade the database to the specified migration revision."""
    log.info(f"Running database downgrade to revision {revision}")
    command.downgrade(alembic_config, revision)


def migrate_db(
    database_url: str,
    revision: str = "head",
    downgrade: bool = False,
) -> None:
    """
    Runs migrations and creates database objects if needed.

    Safe to run on an existing database when upgrading Fides version.
    """
    log.info("Migrating database")
    alembic_config = get_alembic_config(database_url)
    if downgrade:
        downgrade_db(alembic_config, revision)
    else:
        upgrade_db(alembic_config, revision)


def create_db_if_not_exists(database_url: str) -> None:
    """
    Creates a database which does not exist already.
    """
    if not database_exists(database_url):
        create_database(database_url)


def reset_db(database_url: str) -> None:
    """
    Drops all tables/metadata from the database.
    """
    log.info("Resetting database...")
    engine = get_db_engine(database_uri=database_url)
    with engine.connect() as connection:
        log.info("Dropping tables...")
        Base.metadata.drop_all(connection)

        log.info("Dropping excluded tables without models...")
        for table_name in EXCLUDED_TABLES:
            connection.execute(text(f"DROP TABLE IF EXISTS {table_name} CASCADE"))

        log.info("Dropping Alembic table...")
        migration_context = migration.MigrationContext.configure(connection)
        version = migration_context._version  # pylint: disable=protected-access
        if version.exists(connection):
            version.drop(connection)
    log.info("Reset complete.")


def get_db_health(
    database_url: str, db: Session
) -> Tuple[DatabaseHealth, Optional[str]]:
    """Checks if the db is reachable and up-to-date with Alembic migrations."""
    try:
        alembic_config = get_alembic_config(database_url)
        alembic_script_directory = script.ScriptDirectory.from_config(alembic_config)
        context = migration.MigrationContext.configure(db.connection())
        current_revision = context.get_current_revision()
        if (
            context.get_current_revision()
            != alembic_script_directory.get_current_head()
        ):
            db_health: DatabaseHealth = "needs migration"
        else:
            db_health = "healthy"
        return db_health, current_revision
    except Exception as error:  # pylint: disable=broad-except
        error_type = get_full_exception_name(error)
        log.error("Unable to reach the database: {}: {}", error_type, error)
        return ("unhealthy", None)


def seed_db(session: Session) -> None:
    """Load default resources into the database"""
    load_default_resources(session)


def configure_db(database_url: str, revision: Optional[str] = "head") -> None:
    """Set up the db to be used by the app. Creates db if needed and runs migrations."""
    try:
        create_db_if_not_exists(database_url)
        migrate_db(database_url, revision=revision)  # type: ignore[arg-type]

    except InvalidCiphertextError as cipher_error:
        log.error(
            "Unable to configure database due to a decryption error! Check to ensure your `app_encryption_key` has not changed."
        )
        log.opt(exception=True).error(cipher_error)

    except Exception as error:  # pylint: disable=broad-except
        error_type = get_full_exception_name(error)
        log.error("Unable to configure database: {}: {}", error_type, error)
        log.opt(exception=True).error(error)
        raise


def check_missing_migrations(database_url: str) -> None:
    """
    Tries to autogenerate migrations, returns True if a migration
    was generated.
    """
    engine = get_db_engine(database_uri=database_url)
    connection = engine.connect()

    migration_context = migration.MigrationContext.configure(
        connection, opts={"include_object": include_object}
    )
    result = command.autogen.compare_metadata(migration_context, Base.metadata)  # type: ignore[attr-defined]

    if result:
        log.error("Migration needs to be generated!")
        log.error("Detected the following schema differences:")
        for item in result:
            log.error(f"  - {item}")
        raise SystemExit("Migration needs to be generated!")
    log.info("No migrations need to be generated.")


def _get_base_branch_for_comparison() -> str:
    """
    Determine the base branch to compare against for detecting new migrations.

    In GitHub Actions PRs, uses GITHUB_BASE_REF (the PR target branch).
    Otherwise, falls back to origin/main for local development.

    Returns:
        The base branch reference (e.g., 'origin/main', 'origin/develop')
    """
    base_ref = os.environ.get("GITHUB_BASE_REF")
    if base_ref:
        base_branch = f"origin/{base_ref}"
        log.info(f"Comparing against PR target branch: {base_branch}")
    else:
        base_branch = "origin/main"
        log.info(f"No GITHUB_BASE_REF found, comparing against: {base_branch}")
    return base_branch


def _get_new_migration_files(repo_root: str, base_branch: str) -> list[str]:
    """
    Get list of new migration files added in the current branch.

    Files are sorted by the timestamp in their filename to ensure chronological order.
    Migration files follow the pattern: xx_YYYY_MM_DD_HHMM_revision_description.py

    Args:
        repo_root: Absolute path to the git repository root
        base_branch: The base branch to compare against

    Returns:
        List of relative file paths to new migration files, sorted chronologically

    Raises:
        SystemExit: If the migrations directory does not exist
    """
    # Validate that the migrations directory exists
    migrations_dir = (
        Path(repo_root)
        / "src"
        / "fides"
        / "api"
        / "alembic"
        / "migrations"
        / "versions"
    )
    if not migrations_dir.exists():
        log.error(f"Migrations directory not found: {migrations_dir}")
        raise SystemExit(
            f"Migration directory not found at expected path: {migrations_dir}\n"
            "If the migrations directory has been moved, update the path in "
            "fides.api.db.database._get_new_migration_files()"
        )

    migration_pattern = "src/fides/api/alembic/migrations/versions/*.py"
    diff_output = subprocess.check_output(
        [
            "git",
            "diff",
            "--name-only",
            "--diff-filter=A",  # Only added files
            f"{base_branch}...HEAD",
            "--",
            migration_pattern,
        ],
        stderr=subprocess.STDOUT,
        text=True,
        cwd=repo_root,
    ).strip()

    if not diff_output:
        return []

    migration_files = [f for f in diff_output.split("\n") if f]

    # Sort by filename to ensure chronological order
    # Files are named: xx_YYYY_MM_DD_HHMM_revision_description.py
    # Sorting alphabetically will sort them chronologically due to the date format
    migration_files.sort()

    return migration_files


def _parse_migration_revisions(migration_path: Path) -> tuple[str, Optional[str]]:
    """
    Extract revision and down_revision IDs from a migration file.

    Args:
        migration_path: Path to the migration file

    Returns:
        Tuple of (revision, down_revision). down_revision may be None.

    Raises:
        SystemExit: If revision ID cannot be found in the file
    """
    revision = None
    down_revision = None

    with open(migration_path, "r", encoding="utf-8") as f:
        for line in f:
            if line.strip().startswith("revision ="):
                revision = line.split("=")[1].strip().strip("\"'")
            elif line.strip().startswith("down_revision ="):
                down_rev_value = line.split("=")[1].strip().strip("\"'")
                if down_rev_value.lower() != "none":
                    down_revision = down_rev_value

    if not revision:
        log.error(f"Could not find revision ID in {migration_path}")
        raise SystemExit(f"Invalid migration file: {migration_path}")

    return revision, down_revision


def _test_migrations_upgrade_downgrade_sequence(
    alembic_config: Config,
    migrations: list[tuple[str, str, Optional[str]]],
) -> None:
    """
    Test multiple migrations can be upgraded and downgraded successfully in sequence.

    Tests all new migrations together by:
    1. Downgrading incrementally through all migrations (newest to oldest)
    2. Upgrading incrementally back through all migrations (oldest to newest)

    After Phase 2, the database is at the newest migration revision, which should
    match HEAD for the new migrations being tested.

    This approach tests migrations in a realistic sequence and can catch
    interaction issues between migrations.

    Args:
        alembic_config: Alembic configuration
        migrations: List of (migration_name, revision, down_revision) tuples

    Raises:
        SystemExit: If any migration upgrade/downgrade test fails
    """
    if not migrations:
        return

    log.info(f"Testing {len(migrations)} migration(s) in sequence")

    try:
        # Phase 1: Downgrade through all migrations (newest to oldest)
        log.info("Phase 1: Downgrading through all new migrations...")
        for migration_name, revision, down_revision in reversed(migrations):
            if down_revision:
                log.info(f"  Downgrading {migration_name} to {down_revision}...")
                downgrade_db(alembic_config, down_revision)
            else:
                log.info(f"  Skipping {migration_name} (no down_revision)")

        # Phase 2: Upgrade back through all migrations (oldest to newest)
        log.info("Phase 2: Upgrading back through all new migrations...")
        for migration_name, revision, down_revision in migrations:
            log.info(f"  Upgrading {migration_name} to {revision}...")
            upgrade_db(alembic_config, revision)

        log.info(f"✓ All {len(migrations)} migration(s) passed upgrade/downgrade test")

    except Exception as error:
        log.error("Migration upgrade/downgrade sequence test failed!")
        log.error(str(error))
        raise SystemExit("Migration upgrade/downgrade test failed")


def check_new_migrations_upgrade_downgrade(database_url: str) -> None:
    """
    Detects new migrations added in the current branch and tests that they
    can be successfully upgraded and downgraded.

    Uses git to compare against the target/base branch to find new migration files.
    In CI, uses GITHUB_BASE_REF if available, otherwise falls back to origin/main.
    """
    try:
        # Get the repo root directory
        repo_root = subprocess.check_output(
            ["git", "rev-parse", "--show-toplevel"],
            stderr=subprocess.STDOUT,
            text=True,
        ).strip()

        # Determine base branch and get new migration files
        base_branch = _get_base_branch_for_comparison()
        migration_files = _get_new_migration_files(repo_root, base_branch)

        if not migration_files:
            log.info("No new migration files detected in current branch")
            log.info("No new migrations to test.")
            return

        log.info(f"Detected {len(migration_files)} new migration(s) to test")

        # Log the migrations in order
        for i, file_path in enumerate(migration_files, 1):
            log.info(f"  {i}. {Path(file_path).name}")

        alembic_config = get_alembic_config(database_url)

        # Parse all migrations first
        migrations = []
        for file_path in migration_files:
            abs_path = Path(repo_root) / file_path
            revision, down_revision = _parse_migration_revisions(abs_path)
            migrations.append((abs_path.name, revision, down_revision))

        # Test all migrations in sequence
        _test_migrations_upgrade_downgrade_sequence(alembic_config, migrations)

        log.info(
            f"All {len(migration_files)} new migration(s) passed upgrade/downgrade tests."
        )

    except subprocess.CalledProcessError as error:
        # Git command failed - might not be in a git repo or no origin/main
        log.warning(f"Could not detect new migrations via git: {error}")
        log.warning(
            "Skipping migration upgrade/downgrade check (git detection failed)."
        )



"""
Pure utility functions for async DSR operations.

This module contains utility functions with no business logic dependencies.
These are helper functions that can be used across the async DSR system.
"""

from enum import Enum

# Type checking imports
from typing import TYPE_CHECKING, List, cast

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import PrivacyRequestError
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.datasetconfig import DatasetConfig
from fides.api.models.privacy_request.request_task import AsyncTaskType, RequestTask
from fides.api.schemas.saas.saas_config import SaaSRequest

if TYPE_CHECKING:
    from fides.api.service.connectors.query_configs.saas_query_config import (
        SaaSQueryConfig,
    )


def has_async_requests(query_config: "SaaSQueryConfig") -> bool:
    """Check if any read/update/delete requests have async configuration"""
    all_requests: List[SaaSRequest] = cast(
        List[SaaSRequest], query_config.get_read_requests_by_identity()
    )
    masking_request = query_config.get_masking_request()
    if masking_request:
        all_requests.append(masking_request)
    return any(request.async_config is not None for request in all_requests)


def is_polling_continuation(request_task: RequestTask) -> bool:
    """Check if this is a polling continuation (not initial request)"""
    async_type_check = request_task.async_type == AsyncTaskType.polling
    sub_requests_count = len(request_task.sub_requests)

    logger.warning(
        f"is_polling_continuation for task {request_task.id}: "
        f"async_type={request_task.async_type} (polling={async_type_check}), "
        f"sub_requests_count={sub_requests_count}"
    )

    return async_type_check and sub_requests_count > 0


def is_callback_completion(request_task: RequestTask) -> bool:
    """Check if this is a completed callback request"""
    return bool(request_task.callback_succeeded)


def is_async_request(
    request_task: RequestTask, query_config: "SaaSQueryConfig"
) -> bool:
    """
    Check if this is an async request (callback, continuation, or initial async).

    This is the main entry point for async detection.
    """
    return (
        is_callback_completion(request_task)
        or is_polling_continuation(request_task)
        or has_async_requests(query_config)
    )


class AsyncPhase(Enum):
    """Enum representing different phases of async DSR processing."""

    callback_completion = "callback_completion"
    polling_continuation = "polling_continuation"
    initial_async = "initial_async"
    sync = "sync"


def get_async_phase(
    request_task: RequestTask, query_config: "SaaSQueryConfig"
) -> AsyncPhase:
    """
    Classify the phase of async request for routing purposes.

    Returns:
        AsyncPhase enum value representing the current phase:
        - callback_completion: Callback has completed
        - polling_continuation: Continuing an existing polling process
        - initial_async: Initial async request setup
        - sync: Not an async request
    """
    if is_callback_completion(request_task):
        return AsyncPhase.callback_completion
    if is_polling_continuation(request_task):
        return AsyncPhase.polling_continuation
    if has_async_requests(query_config):
        return AsyncPhase.initial_async
    return AsyncPhase.sync


def get_connection_config_from_task(
    db: Session, request_task: RequestTask
) -> ConnectionConfig:
    """
    Get ConnectionConfig from a RequestTask.

    This utility function retrieves the connection configuration
    associated with a request task by looking up the dataset configuration.
    """
    dataset_config = DatasetConfig.filter(
        db=db,
        conditions=(DatasetConfig.fides_key == request_task.dataset_name),
    ).first()
    if not dataset_config:
        raise PrivacyRequestError(
            f"DatasetConfig with fides_key {request_task.dataset_name} not found."
        )

    connection_config = ConnectionConfig.get(
        db=db, object_id=dataset_config.connection_config_id
    )
    if not connection_config:
        raise PrivacyRequestError(
            f"ConnectionConfig with id {dataset_config.connection_config_id} not found."
        )

    return connection_config



"""Update user_assigned_data_uses so it's nullable and defaults to none

Revision ID: 9caf76161e55
Revises: 918aefc950c9
Create Date: 2025-09-29 14:14:37.837616

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "9caf76161e55"
down_revision = "918aefc950c9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column(
        "stagedresource",
        "user_assigned_data_uses",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=True,
        server_default=None,
        default=None,
        existing_server_default=sa.text("'{}'::character varying[]"),
    )

    # Update web monitor resources with empty array to null
    op.execute(
        """
        UPDATE stagedresource
        SET user_assigned_data_uses = NULL
        WHERE
          resource_type IN ('Cookie', 'Browser request', 'Image', 'iFrame', 'Javascript tag')
          AND user_assigned_data_uses = '{}'
        """
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute(
        """
        UPDATE stagedresource
        SET user_assigned_data_uses = '{}'
        WHERE user_assigned_data_uses IS NULL
        """
    )

    op.alter_column(
        "stagedresource",
        "user_assigned_data_uses",
        existing_type=postgresql.ARRAY(sa.VARCHAR()),
        nullable=False,
        server_default=sa.text("'{}'::character varying[]"),
        existing_server_default=None,
        default=dict,
    )
    # ### end Alembic commands ###



from io import BytesIO
from typing import List
from zipfile import BadZipFile, ZipFile

from fastapi import Body, Depends, HTTPException
from fastapi.params import Security
from fastapi.responses import JSONResponse, Response
from loguru import logger
from sqlalchemy.orm import Session
from starlette.status import HTTP_404_NOT_FOUND

from fides.api.api import deps
from fides.api.oauth.utils import verify_oauth_client
from fides.api.schemas.saas.connector_template import ConnectorTemplateListResponse
from fides.api.service.connectors.saas.connector_registry_service import (
    ConnectorRegistry,
    CustomConnectorTemplateLoader,
)
from fides.api.util.api_router import APIRouter
from fides.common.api.scope_registry import (
    CONNECTOR_TEMPLATE_READ,
    CONNECTOR_TEMPLATE_REGISTER,
)
from fides.common.api.v1.urn_registry import (
    CONNECTOR_TEMPLATES,
    CONNECTOR_TEMPLATES_CONFIG,
    CONNECTOR_TEMPLATES_DATASET,
    CONNECTOR_TEMPLATES_REGISTER,
    REGISTER_CONNECTOR_TEMPLATE,
    V1_URL_PREFIX,
)

router = APIRouter(tags=["Connector Templates"], prefix=V1_URL_PREFIX)


@router.get(
    CONNECTOR_TEMPLATES,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTOR_TEMPLATE_READ])],
    response_model=List[ConnectorTemplateListResponse],
)
def get_all_connector_templates() -> List[ConnectorTemplateListResponse]:
    """
    Returns a list of all available connector templates with summary information.

    Each template includes:
    - **type**: The unique identifier for the connector (used in other endpoints)
    - **name**: Human-readable name of the connector
    - **supported_actions**: List of actions this connector supports (e.g., access, erasure)
    - **category**: Optional category classification for the connector
    - **custom**: Whether this is a custom connector (true) or built-in (false)
    """
    logger.info("Retrieving all connector templates")
    return ConnectorRegistry.get_all_connector_templates_summary()


@router.post(
    CONNECTOR_TEMPLATES_REGISTER,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTOR_TEMPLATE_REGISTER])],
)
def register_connector_template(
    file: bytes = Body(..., media_type="application/zip"),
    db: Session = Depends(deps.get_db),
) -> JSONResponse:
    """
    Registers a custom connector template from a zip file uploaded by the user.

    The endpoint performs the following steps:
    1. Validates the uploaded file is a proper zip file.
    2. Uses the CustomConnectorTemplateLoader to validate, register, and save the template to the database.

    If the uploaded file is not a valid zip file or there are any validation errors
    when creating the ConnectorTemplates an HTTP 400 status code with error details is returned.
    """
    try:
        with ZipFile(BytesIO(file), "r") as zip_file:
            CustomConnectorTemplateLoader.save_template(db=db, zip_file=zip_file)
    except BadZipFile:
        raise HTTPException(status_code=400, detail="Invalid zip file")
    except Exception as exc:
        logger.exception("Error loading connector template from zip file.")
        raise HTTPException(status_code=400, detail=str(exc))

    return JSONResponse(
        content={"message": "Connector template successfully registered."}
    )


@router.post(
    REGISTER_CONNECTOR_TEMPLATE,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTOR_TEMPLATE_REGISTER])],
    deprecated=True,
)
def register_custom_connector_template_deprecated(
    file: bytes = Body(..., media_type="application/zip"),
    db: Session = Depends(deps.get_db),
) -> JSONResponse:
    """
    **DEPRECATED**: Use `POST /connector-templates/register` instead.

    Registers a custom connector template from a zip file uploaded by the user.
    This endpoint is maintained for backward compatibility only.
    """
    return register_connector_template(file=file, db=db)


@router.get(
    CONNECTOR_TEMPLATES_CONFIG,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTOR_TEMPLATE_READ])],
)
def get_connector_template_config(
    connector_template_type: str,
) -> Response:
    """
    Retrieves the SaaS config YAML for a connector template by its type.

    The `connector_template_type` parameter comes from the `type` field
    returned by the `GET /connector-templates` endpoint.

    Returns the raw YAML configuration that can be used to understand
    or customize the connector template.
    """
    logger.info("Finding connector template with type '{}'", connector_template_type)
    template = ConnectorRegistry.get_connector_template(connector_template_type)

    if not template:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No connector template found with type '{connector_template_type}'",
        )

    return Response(
        content=template.config,
        media_type="text/yaml",
    )


@router.get(
    CONNECTOR_TEMPLATES_DATASET,
    dependencies=[Security(verify_oauth_client, scopes=[CONNECTOR_TEMPLATE_READ])],
)
def get_connector_template_dataset(
    connector_template_type: str,
) -> Response:
    """
    Retrieves the dataset YAML for a connector template by its type.

    The `connector_template_type` parameter comes from the `type` field
    returned by the `GET /connector-templates` endpoint.

    Returns the raw dataset YAML configuration that defines the data structure
    for the connector template.
    """
    logger.info(
        "Finding connector template dataset with type '{}'", connector_template_type
    )
    template = ConnectorRegistry.get_connector_template(connector_template_type)

    if not template:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No connector template found with type '{connector_template_type}'",
        )

    return Response(
        content=template.dataset,
        media_type="text/yaml",
    )



"""added duplicate group table

Revision ID: 80d28dea3b6b
Revises: c09e76282dd1
Create Date: 2025-10-29 16:59:09.551676

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "80d28dea3b6b"
down_revision = "c09e76282dd1"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "duplicate_group",
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("id", postgresql.UUID(as_uuid=True), nullable=False),
        sa.Column("rule_version", sa.Text(), nullable=False),
        sa.Column("is_active", sa.Boolean(), server_default="t", nullable=False),
        sa.Column("dedup_key", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.drop_index(
        "ix_privacyrequest_duplicate_request_group_id", table_name="privacyrequest"
    )
    op.drop_column("privacyrequest", "duplicate_request_group_id")
    op.add_column(
        "privacyrequest",
        sa.Column(
            "duplicate_request_group_id", postgresql.UUID(as_uuid=True), nullable=True
        ),
    )
    op.create_foreign_key(
        "fk_privacyrequest_duplicate_group",
        "privacyrequest",
        "duplicate_group",
        ["duplicate_request_group_id"],
        ["id"],
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "fk_privacyrequest_duplicate_group", "privacyrequest", type_="foreignkey"
    )
    op.drop_column("privacyrequest", "duplicate_request_group_id")
    op.add_column(
        "privacyrequest",
        sa.Column("duplicate_request_group_id", sa.String(), nullable=True),
    )
    op.create_index(
        "ix_privacyrequest_duplicate_request_group_id",
        "privacyrequest",
        ["duplicate_request_group_id"],
        unique=False,
    )
    op.drop_table("duplicate_group")
    # ### end Alembic commands ###
